"""Differential operators."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.argument import Argument, Coargument
from ufl.checks import is_cellwise_constant
from ufl.coefficient import Coefficient
from ufl.constantvalue import Zero
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.expr import Expr
from ufl.core.operator import Operator
from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type
from ufl.domain import extract_unique_domain, find_geometric_dimension
from ufl.exprcontainers import ExprList, ExprMapping
from ufl.form import BaseForm
from ufl.precedence import parstr
from ufl.variable import Variable

# --- Basic differentiation objects ---


@ufl_type(is_abstract=True, is_differential=True)
class Derivative(Operator):
    """Base class for all derivative types."""

    __slots__ = ()

    def __init__(self, operands):
        """Initalise."""
        Operator.__init__(self, operands)


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class CoefficientDerivative(Derivative):
    """Derivative of form integrand w.r.t. the degrees of freedom in a discrete Coefficient."""

    __slots__ = ()

    def __new__(cls, integrand, coefficients, arguments, coefficient_derivatives):
        """Create a new CoefficientDerivative."""
        if not isinstance(coefficients, ExprList):
            raise ValueError("Expecting ExprList instance with Coefficients.")
        if not isinstance(arguments, ExprList):
            raise ValueError("Expecting ExprList instance with Arguments.")
        if not isinstance(coefficient_derivatives, ExprMapping):
            raise ValueError("Expecting ExprMapping for coefficient derivatives.")
        if isinstance(integrand, Zero):
            return integrand
        return Derivative.__new__(cls)

    def __init__(self, integrand, coefficients, arguments, coefficient_derivatives):
        """Initalise."""
        if not isinstance(coefficient_derivatives, ExprMapping):
            coefficient_derivatives = ExprMapping(coefficient_derivatives)
        Derivative.__init__(self, (integrand, coefficients, arguments, coefficient_derivatives))

    def __str__(self):
        """Format as a string."""
        return "d/dfj { %s }, with fh=%s, dfh/dfj = %s, and coefficient derivatives %s" % (
            self.ufl_operands[0],
            self.ufl_operands[1],
            self.ufl_operands[2],
            self.ufl_operands[3],
        )


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class CoordinateDerivative(CoefficientDerivative):
    """Derivative of the integrand of a form w.r.t. the SpatialCoordinates."""

    __slots__ = ()

    def __str__(self):
        """Format as a string."""
        return "d/dfj { %s }, with fh=%s, dfh/dfj = %s, and coordinate derivatives %s" % (
            self.ufl_operands[0],
            self.ufl_operands[1],
            self.ufl_operands[2],
            self.ufl_operands[3],
        )


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class BaseFormDerivative(CoefficientDerivative, BaseForm):
    """Derivative of a base form w.r.t the degrees of freedom in a discrete Coefficient."""

    _ufl_noslots_ = True

    def __init__(self, base_form, coefficients, arguments, coefficient_derivatives):
        """Initalise."""
        CoefficientDerivative.__init__(
            self, base_form, coefficients, arguments, coefficient_derivatives
        )
        BaseForm.__init__(self)

    def _analyze_form_arguments(self):
        """Collect the arguments of the corresponding BaseForm."""
        from ufl.algorithms.analysis import extract_coefficients, extract_type

        base_form, _, arguments, _ = self.ufl_operands

        def arg_type(x):
            if isinstance(x, BaseForm):
                return Coargument
            return Argument

        # Each derivative arguments can either be a:
        # - `ufl.BaseForm`: if it contains a `ufl.Coargument`
        # - or a `ufl.Expr`: if it contains a `ufl.Argument`
        # When a `Coargument` is encountered, it is treated as an
        # argument (i.e. as V* -> V* and not V* x V -> R) and should
        # result in one single argument (in the dual space).
        base_form_args = base_form.arguments() + tuple(
            arg for a in arguments.ufl_operands for arg in extract_type(a, arg_type(a))
        )
        # BaseFormDerivative's arguments don't necessarily contain BaseArgument objects only
        # -> e.g. `derivative(u ** 2, u, u)` with `u` a Coefficient.
        base_form_coeffs = base_form.coefficients() + tuple(
            arg for a in arguments.ufl_operands for arg in extract_coefficients(a)
        )
        # Reconstruct arguments for correct numbering
        self._arguments = tuple(
            type(arg)(arg.ufl_function_space(), arg.number(), arg.part()) for arg in base_form_args
        )
        self._coefficients = base_form_coeffs


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class BaseFormCoordinateDerivative(BaseFormDerivative, CoordinateDerivative):
    """Derivative of a base form w.r.t. the SpatialCoordinates."""

    _ufl_noslots_ = True

    def __init__(self, base_form, coefficients, arguments, coefficient_derivatives):
        """Initalise."""
        BaseFormDerivative.__init__(
            self, base_form, coefficients, arguments, coefficient_derivatives
        )


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class BaseFormOperatorDerivative(BaseFormDerivative, BaseFormOperator):
    """Derivative of a base form operator w.r.t the degrees of freedom in a discrete Coefficient."""

    _ufl_noslots_ = True

    # BaseFormOperatorDerivative is only needed because of a different
    # differentiation procedure for BaseformOperator objects.
    def __init__(self, base_form, coefficients, arguments, coefficient_derivatives):
        """Initalise."""
        BaseFormDerivative.__init__(
            self, base_form, coefficients, arguments, coefficient_derivatives
        )
        self._argument_slots = base_form._argument_slots

    # Enforce Operator reconstruction as Operator is a parent class of
    # both: BaseFormDerivative and BaseFormOperator.
    # Therefore the latter overwrites Operator reconstruction and we would have:
    #   -> BaseFormOperatorDerivative._ufl_expr_reconstruct_ =
    #   BaseFormOperator._ufl_expr_reconstruct_
    _ufl_expr_reconstruct_ = Operator._ufl_expr_reconstruct_
    # Set __repr__
    __repr__ = Operator.__repr__

    def argument_slots(self, outer_form=False):
        """Return a tuple of expressions containing argument and coefficient based expressions."""
        from ufl.algorithms.analysis import extract_arguments

        base_form, _, arguments, _ = self.ufl_operands
        argument_slots = base_form.argument_slots(outer_form) + tuple(
            arg for a in arguments for arg in extract_arguments(a)
        )
        return argument_slots


@ufl_type(num_ops=4, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class BaseFormOperatorCoordinateDerivative(BaseFormOperatorDerivative, CoordinateDerivative):
    """Derivative of a base form operator w.r.t. the SpatialCoordinates."""

    _ufl_noslots_ = True

    def __init__(self, base_form, coefficients, arguments, coefficient_derivatives):
        """Initalise."""
        BaseFormOperatorDerivative.__init__(
            self, base_form, coefficients, arguments, coefficient_derivatives
        )


@ufl_type(num_ops=2)
class VariableDerivative(Derivative):
    """Variable Derivative."""

    __slots__ = (
        "ufl_shape",
        "ufl_free_indices",
        "ufl_index_dimensions",
    )

    def __new__(cls, f, v):
        """Create a new VariableDerivative."""
        # Checks
        if not isinstance(f, Expr):
            raise ValueError("Expecting an Expr in VariableDerivative.")
        if not isinstance(v, (Variable, Coefficient)):
            raise ValueError("Expecting a Variable in VariableDerivative.")
        if v.ufl_free_indices:
            raise ValueError("Differentiation variable cannot have free indices.")

        # Simplification
        # Return zero if expression is trivially independent of variable
        if f._ufl_is_terminal_ and f != v:
            return Zero(f.ufl_shape + v.ufl_shape, f.ufl_free_indices, f.ufl_index_dimensions)

        # Construction
        return Derivative.__new__(cls)

    def __init__(self, f, v):
        """Initalise."""
        Derivative.__init__(self, (f, v))
        self.ufl_free_indices = f.ufl_free_indices
        self.ufl_index_dimensions = f.ufl_index_dimensions
        self.ufl_shape = f.ufl_shape + v.ufl_shape

    def __str__(self):
        """Format as a string."""
        if isinstance(self.ufl_operands[0], Terminal):
            return "d%s/d[%s]" % (self.ufl_operands[0], self.ufl_operands[1])
        return "d/d[%s] %s" % (self.ufl_operands[1], parstr(self.ufl_operands[0], self))


# --- Compound differentiation objects ---


@ufl_type(is_abstract=True)
class CompoundDerivative(Derivative):
    """Base class for all compound derivative types."""

    __slots__ = ()

    def __init__(self, operands):
        """Initalise."""
        Derivative.__init__(self, operands)


@ufl_type(num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True)
class Grad(CompoundDerivative):
    """Grad."""

    __slots__ = ("_dim",)

    def __new__(cls, f):
        """Create a new Grad."""
        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            dim = find_geometric_dimension(f)
            return Zero(f.ufl_shape + (dim,), f.ufl_free_indices, f.ufl_index_dimensions)
        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))
        self._dim = find_geometric_dimension(f)

    def _ufl_expr_reconstruct_(self, op):
        """Return a new object of the same type with new operands."""
        if is_cellwise_constant(op):
            if op.ufl_shape != self.ufl_operands[0].ufl_shape:
                raise ValueError("Operand shape mismatch in Grad reconstruct.")
            if self.ufl_operands[0].ufl_free_indices != op.ufl_free_indices:
                raise ValueError("Free index mismatch in Grad reconstruct.")
            return Zero(self.ufl_shape, self.ufl_free_indices, self.ufl_index_dimensions)
        return self._ufl_class_(op)

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Get child from mapping and return the component asked for."""
        component, i = component[:-1], component[-1]
        derivatives = derivatives + (i,)
        result = self.ufl_operands[0].evaluate(
            x, mapping, component, index_values, derivatives=derivatives
        )
        return result

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape + (self._dim,)

    def __str__(self):
        """Format as a string."""
        return "grad(%s)" % self.ufl_operands[0]


@ufl_type(
    num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True, is_in_reference_frame=True
)
class ReferenceGrad(CompoundDerivative):
    """Reference grad."""

    __slots__ = ("_dim",)

    def __new__(cls, f):
        """Create a new ReferenceGrad."""
        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            dim = extract_unique_domain(f).topological_dimension()
            return Zero(f.ufl_shape + (dim,), f.ufl_free_indices, f.ufl_index_dimensions)
        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))
        self._dim = extract_unique_domain(f).topological_dimension()

    def _ufl_expr_reconstruct_(self, op):
        """Return a new object of the same type with new operands."""
        if is_cellwise_constant(op):
            if op.ufl_shape != self.ufl_operands[0].ufl_shape:
                raise ValueError("Operand shape mismatch in ReferenceGrad reconstruct.")
            if self.ufl_operands[0].ufl_free_indices != op.ufl_free_indices:
                raise ValueError("Free index mismatch in ReferenceGrad reconstruct.")
            return Zero(self.ufl_shape, self.ufl_free_indices, self.ufl_index_dimensions)
        return self._ufl_class_(op)

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Get child from mapping and return the component asked for."""
        component, i = component[:-1], component[-1]
        derivatives = derivatives + (i,)
        result = self.ufl_operands[0].evaluate(
            x, mapping, component, index_values, derivatives=derivatives
        )
        return result

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape + (self._dim,)

    def __str__(self):
        """Format as a string."""
        return "reference_grad(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True)
class Div(CompoundDerivative):
    """Div."""

    __slots__ = ()

    def __new__(cls, f):
        """Create a new Div."""
        if f.ufl_free_indices:
            raise ValueError("Free indices in the divergence argument is not allowed.")

        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            return Zero(f.ufl_shape[:-1])  # No free indices asserted above

        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape[:-1]

    def __str__(self):
        """Format as a string."""
        return "div(%s)" % self.ufl_operands[0]


@ufl_type(
    num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True, is_in_reference_frame=True
)
class ReferenceDiv(CompoundDerivative):
    """Reference divergence."""

    __slots__ = ()

    def __new__(cls, f):
        """Create a new ReferenceDiv."""
        if f.ufl_free_indices:
            raise ValueError("Free indices in the divergence argument is not allowed.")

        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            return Zero(f.ufl_shape[:-1])  # No free indices asserted above

        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape[:-1]

    def __str__(self):
        """Format as a string."""
        return "reference_div(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_indices_from_operand=0)
class NablaGrad(CompoundDerivative):
    """Nabla grad."""

    __slots__ = ("_dim",)

    def __new__(cls, f):
        """Create a new NablaGrad."""
        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            dim = find_geometric_dimension(f)
            return Zero((dim,) + f.ufl_shape, f.ufl_free_indices, f.ufl_index_dimensions)
        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))
        self._dim = find_geometric_dimension(f)

    def _ufl_expr_reconstruct_(self, op):
        """Return a new object of the same type with new operands."""
        if is_cellwise_constant(op):
            if op.ufl_shape != self.ufl_operands[0].ufl_shape:
                raise ValueError("Operand shape mismatch in NablaGrad reconstruct.")
            if self.ufl_operands[0].ufl_free_indices != op.ufl_free_indices:
                raise ValueError("Free index mismatch in NablaGrad reconstruct.")
            return Zero(self.ufl_shape, self.ufl_free_indices, self.ufl_index_dimensions)
        return self._ufl_class_(op)

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return (self._dim,) + self.ufl_operands[0].ufl_shape

    def __str__(self):
        """Format as a string."""
        return "nabla_grad(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_indices_from_operand=0)
class NablaDiv(CompoundDerivative):
    """Nabla div."""

    __slots__ = ()

    def __new__(cls, f):
        """Create a new NablaDiv."""
        if f.ufl_free_indices:
            raise ValueError("Free indices in the divergence argument is not allowed.")

        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            return Zero(f.ufl_shape[1:])  # No free indices asserted above

        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape[1:]

    def __str__(self):
        """Format as a string."""
        return "nabla_div(%s)" % self.ufl_operands[0]


_curl_shapes = {(): (2,), (2,): (), (3,): (3,)}


@ufl_type(num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True)
class Curl(CompoundDerivative):
    """Compound derivative."""

    __slots__ = ("ufl_shape",)

    def __new__(cls, f):
        """Create a new CompoundDerivative."""
        # Validate input
        sh = f.ufl_shape
        if f.ufl_shape not in ((), (2,), (3,)):
            raise ValueError("Expecting a scalar, 2D vector or 3D vector.")
        if f.ufl_free_indices:
            raise ValueError("Free indices in the curl argument is not allowed.")

        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            sh = {(): (2,), (2,): (), (3,): (3,)}[sh]
            return Zero(sh)  # No free indices asserted above
        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))
        self.ufl_shape = _curl_shapes[f.ufl_shape]

    def __str__(self):
        """Format as a string."""
        return "curl(%s)" % self.ufl_operands[0]


@ufl_type(
    num_ops=1, inherit_indices_from_operand=0, is_terminal_modifier=True, is_in_reference_frame=True
)
class ReferenceCurl(CompoundDerivative):
    """Reference curl."""

    __slots__ = ("ufl_shape",)

    def __new__(cls, f):
        """Create a new ReferenceCurl."""
        # Validate input
        sh = f.ufl_shape
        if f.ufl_shape not in ((), (2,), (3,)):
            raise ValueError("Expecting a scalar, 2D vector or 3D vector.")
        if f.ufl_free_indices:
            raise ValueError("Free indices in the curl argument is not allowed.")

        # Return zero if expression is trivially constant
        if is_cellwise_constant(f):
            sh = {(): (2,), (2,): (), (3,): (3,)}[sh]
            return Zero(sh)  # No free indices asserted above
        return CompoundDerivative.__new__(cls)

    def __init__(self, f):
        """Initalise."""
        CompoundDerivative.__init__(self, (f,))
        self.ufl_shape = _curl_shapes[f.ufl_shape]

    def __str__(self):
        """Format as a string."""
        return "reference_curl(%s)" % self.ufl_operands[0]



"""The Integral class."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009
# Modified by Massimiliano Leoni, 2016.

import ufl
from ufl.checks import is_python_scalar, is_scalar_constant_expression
from ufl.core.expr import Expr
from ufl.protocols import id_or_none

# Export list for ufl.classes
__all_classes__ = ["Integral"]


class Integral(object):
    """An integral over a single domain."""

    __slots__ = (
        "_integrand",
        "_integral_type",
        "_ufl_domain",
        "_subdomain_id",
        "_metadata",
        "_subdomain_data",
    )

    def __init__(self, integrand, integral_type, domain, subdomain_id, metadata, subdomain_data):
        """Initialise."""
        if not isinstance(integrand, Expr):
            raise ValueError("Expecting integrand to be an Expr instance.")
        self._integrand = integrand
        self._integral_type = integral_type
        self._ufl_domain = domain
        self._subdomain_id = subdomain_id
        self._metadata = metadata
        self._subdomain_data = subdomain_data

    def reconstruct(
        self,
        integrand=None,
        integral_type=None,
        domain=None,
        subdomain_id=None,
        metadata=None,
        subdomain_data=None,
    ):
        """Construct a new Integral object with some properties replaced with new values.

        Example:
            <a = Integral instance>
            b = a.reconstruct(expand_compounds(a.integrand()))
            c = a.reconstruct(metadata={'quadrature_degree':2})
        """
        if integrand is None:
            integrand = self.integrand()
        if integral_type is None:
            integral_type = self.integral_type()
        if domain is None:
            domain = self.ufl_domain()
        if subdomain_id is None:
            subdomain_id = self.subdomain_id()
        if metadata is None:
            metadata = self.metadata()
        if subdomain_data is None:
            subdomain_data = self._subdomain_data
        return Integral(integrand, integral_type, domain, subdomain_id, metadata, subdomain_data)

    def integrand(self):
        """Return the integrand expression, which is an ``Expr`` instance."""
        return self._integrand

    def integral_type(self):
        """Return the domain type of this integral."""
        return self._integral_type

    def ufl_domain(self):
        """Return the integration domain of this integral."""
        return self._ufl_domain

    def subdomain_id(self):
        """Return the subdomain id of this integral."""
        return self._subdomain_id

    def metadata(self):
        """Return the compiler metadata this integral has been annotated with."""
        return self._metadata

    def subdomain_data(self):
        """Return the domain data of this integral."""
        return self._subdomain_data

    def __neg__(self):
        """Negate."""
        return self.reconstruct(-self._integrand)

    def __mul__(self, scalar):
        """Multiply."""
        if not is_python_scalar(scalar):
            raise ValueError("Cannot multiply an integral with non-constant values.")
        return self.reconstruct(scalar * self._integrand)

    def __rmul__(self, scalar):
        """Multiply."""
        if not is_scalar_constant_expression(scalar):
            raise ValueError(
                "An integral can only be multiplied by a globally constant scalar expression."
            )
        return self.reconstruct(scalar * self._integrand)

    def __str__(self):
        """Format as a string."""
        fmt = "{ %s } * %s(%s[%s], %s)"
        mname = ufl.measure.integral_type_to_measure_name[self._integral_type]
        s = fmt % (self._integrand, mname, self._ufl_domain, self._subdomain_id, self._metadata)
        return s

    def __repr__(self):
        """Representation."""
        return (
            f"Integral({self._integrand!r}, {self._integral_type!r}, {self._ufl_domain!r}, "
            f"{self._subdomain_id!r}, {self._metadata!r}, {self._subdomain_data!r})"
        )

    def __eq__(self, other):
        """Check equality."""
        return (
            isinstance(other, Integral)
            and self._integral_type == other._integral_type
            and self._ufl_domain == other._ufl_domain
            and self._subdomain_id == other._subdomain_id
            and self._integrand == other._integrand
            and self._metadata == other._metadata
            and id_or_none(self._subdomain_data) == id_or_none(other._subdomain_data)
        )

    def __hash__(self):
        """Hash."""
        # Assuming few collisions by ignoring hash(self._metadata) (a
        # dict is not hashable but we assume it is immutable in
        # practice)
        hashdata = (
            hash(self._integrand),
            self._integral_type,
            hash(self._ufl_domain),
            self._subdomain_id,
            id_or_none(self._subdomain_data),
        )
        return hash(hashdata)



"""Utility algorithms for inspection of and information extraction from UFL objects."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010.
# Modified by Johan Hake, 2010.

from itertools import chain

from ufl.algorithms.traversal import iter_expressions
from ufl.argument import BaseArgument, Coargument
from ufl.coefficient import BaseCoefficient
from ufl.constant import Constant
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.terminal import Terminal
from ufl.corealg.traversal import traverse_unique_terminals, unique_pre_traversal
from ufl.form import BaseForm, Form
from ufl.utils.sorting import sorted_by_count, topological_sorting

# TODO: Some of these can possibly be optimised by implementing
# inlined stack based traversal algorithms


def _sorted_by_number_and_part(seq):
    """Sort items by number and part."""
    return sorted(seq, key=lambda x: (x.number(), x.part()))


def unique_tuple(objects):
    """Return tuple of unique objects, preserving initial ordering."""
    unique_objects = []
    handled = set()
    for obj in objects:
        if obj not in handled:
            handled.add(obj)
            unique_objects.append(obj)
    return tuple(unique_objects)


# --- Utilities to extract information from an expression ---


def extract_type(a, ufl_types):
    """Build a set of all objects found in a whose class is in ufl_types.

    Args:
        a: A BaseForm, Integral or Expr
        ufl_types: A list of UFL types

    Returns:
        All objects found in a whose class is in ufl_type
    """
    if not isinstance(ufl_types, (list, tuple)):
        ufl_types = (ufl_types,)

    if all(t is not BaseFormOperator for t in ufl_types):
        remove_base_form_ops = True
        ufl_types += (BaseFormOperator,)
    else:
        remove_base_form_ops = False

    # BaseForms that aren't forms or base form operators
    # only contain arguments & coefficients
    if isinstance(a, BaseForm) and not isinstance(a, (Form, BaseFormOperator)):
        objects = set()
        arg_types = tuple(t for t in ufl_types if issubclass(t, BaseArgument))
        if arg_types:
            objects.update([e for e in a.arguments() if isinstance(e, arg_types)])
        coeff_types = tuple(t for t in ufl_types if issubclass(t, BaseCoefficient))
        if coeff_types:
            objects.update([e for e in a.coefficients() if isinstance(e, coeff_types)])
        return objects

    if all(issubclass(t, Terminal) for t in ufl_types):
        # Optimization
        objects = set(
            o
            for e in iter_expressions(a)
            for o in traverse_unique_terminals(e)
            if any(isinstance(o, t) for t in ufl_types)
        )
    else:
        objects = set(
            o
            for e in iter_expressions(a)
            for o in unique_pre_traversal(e)
            if any(isinstance(o, t) for t in ufl_types)
        )

    # Need to extract objects contained in base form operators whose
    # type is in ufl_types
    base_form_ops = set(e for e in objects if isinstance(e, BaseFormOperator))
    ufl_types_no_args = tuple(t for t in ufl_types if not issubclass(t, BaseArgument))
    base_form_objects = ()
    for o in base_form_ops:
        # This accounts for having BaseFormOperator in Forms: if N is a BaseFormOperator
        # `N(u; v*) * v * dx` <=> `action(v1 * v * dx, N(...; v*))`
        # where `v`, `v1` are `Argument`s and `v*` a `Coargument`.
        for ai in tuple(arg for arg in o.argument_slots(isinstance(a, Form))):
            # Extracting BaseArguments of an object of which a
            # Coargument is an argument, then we just return the dual
            # argument of the Coargument and not its primal argument.
            if isinstance(ai, Coargument):
                new_types = tuple(Coargument if t is BaseArgument else t for t in ufl_types)
                base_form_objects += tuple(extract_type(ai, new_types))
            else:
                base_form_objects += tuple(extract_type(ai, ufl_types))
        # Look for BaseArguments in BaseFormOperator's argument slots
        # only since that's where they are by definition. Don't look
        # into operands, which is convenient for external operator
        # composition, e.g. N1(N2; v*) where N2 is seen as an operator
        # and not a form.
        slots = o.ufl_operands
        for ai in slots:
            base_form_objects += tuple(extract_type(ai, ufl_types_no_args))
    objects.update(base_form_objects)

    # `Remove BaseFormOperator` objects if there were initially not in `ufl_types`
    if remove_base_form_ops:
        objects -= base_form_ops
    return objects


def has_type(a, ufl_type):
    """Return if an object of class ufl_type or a subclass can be found in a.

    Args:
        a: A BaseForm, Integral or Expr
        ufl_type: A UFL type

    Returns:
        Whether an object of class ufl_type can be found in a
    """
    if issubclass(ufl_type, Terminal):
        # Optimization
        traversal = traverse_unique_terminals
    else:
        traversal = unique_pre_traversal
    return any(isinstance(o, ufl_type) for e in iter_expressions(a) for o in traversal(e))


def has_exact_type(a, ufl_type):
    """Return if an object of class ufl_type can be found in a.

    Args:
        a: A BaseForm, Integral or Expr
        ufl_type: A UFL type

    Returns:
        Whether an object of class ufl_type can be found in a
    """
    tc = ufl_type._ufl_typecode_
    if issubclass(ufl_type, Terminal):
        # Optimization
        traversal = traverse_unique_terminals
    else:
        traversal = unique_pre_traversal
    return any(o._ufl_typecode_ == tc for e in iter_expressions(a) for o in traversal(e))


def extract_arguments(a):
    """Build a sorted list of all arguments in a.

    Args:
        a: A BaseForm, Integral or Expr
    """
    return _sorted_by_number_and_part(extract_type(a, BaseArgument))


def extract_coefficients(a):
    """Build a sorted list of all coefficients in a.

    Args:
        a: A BaseForm, Integral or Expr
    """
    return sorted_by_count(extract_type(a, BaseCoefficient))


def extract_constants(a):
    """Build a sorted list of all constants in a.

    Args:
        a: A BaseForm, Integral or Expr
    """
    return sorted_by_count(extract_type(a, Constant))


def extract_base_form_operators(a):
    """Build a sorted list of all base form operators in a.

    Args:
        a: A BaseForm, Integral or Expr
    """
    return sorted_by_count(extract_type(a, BaseFormOperator))


def extract_arguments_and_coefficients(a):
    """Build two sorted lists of all arguments and coefficients in a.

    This function is faster than extract_arguments + extract_coefficients
    for large forms, and has more validation built in.

    Args:
        a: A BaseForm, Integral or Expr
    """
    # Extract lists of all BaseArgument and BaseCoefficient instances
    base_coeff_and_args = extract_type(a, (BaseArgument, BaseCoefficient))
    arguments = [f for f in base_coeff_and_args if isinstance(f, BaseArgument)]
    coefficients = [f for f in base_coeff_and_args if isinstance(f, BaseCoefficient)]

    # Build number,part: instance mappings, should be one to one
    bfnp = dict((f, (f.number(), f.part())) for f in arguments)
    if len(bfnp) != len(set(bfnp.values())):
        raise ValueError(
            "Found different Arguments with same number and part.\n"
            "Did you combine test or trial functions from different spaces?\n"
            "The Arguments found are:\n" + "\n".join(f"  {a}" for a in arguments)
        )

    # Build count: instance mappings, should be one to one
    fcounts = dict((f, f.count()) for f in coefficients)
    if len(fcounts) != len(set(fcounts.values())):
        raise ValueError(
            "Found different coefficients with same counts.\n"
            "The arguments found are:\n" + "\n".join(f"  {c}" for c in coefficients)
        )

    # Passed checks, so we can safely sort the instances by count
    arguments = _sorted_by_number_and_part(arguments)
    coefficients = sorted_by_count(coefficients)

    return arguments, coefficients


def extract_elements(form):
    """Build sorted tuple of all elements used in form."""
    args = chain(*extract_arguments_and_coefficients(form))
    return tuple(f.ufl_element() for f in args)


def extract_unique_elements(form):
    """Build sorted tuple of all unique elements used in form."""
    return unique_tuple(extract_elements(form))


def extract_sub_elements(elements):
    """Build sorted tuple of all sub elements (including parent element)."""
    sub_elements = tuple(chain(*[e.sub_elements for e in elements]))
    if not sub_elements:
        return tuple(elements)
    return tuple(elements) + extract_sub_elements(sub_elements)


def sort_elements(elements):
    """Sort elements.

    A sort is performed so that any sub elements appear before the
    corresponding mixed elements. This is useful when sub elements
    need to be defined before the corresponding mixed elements.

    The ordering is based on sorting a directed acyclic graph.
    """
    # Set nodes
    nodes = list(elements)

    # Set edges
    edges = dict((node, []) for node in nodes)
    for element in elements:
        for sub_element in element.sub_elements:
            edges[element].append(sub_element)

    # Sort graph
    sorted_elements = topological_sorting(nodes, edges)

    # Reverse list of elements
    sorted_elements.reverse()

    return sorted_elements



from ufl import Measure, Mesh, triangle


class MockMesh:
    def __init__(self, ufl_id):
        self._ufl_id = ufl_id

    def ufl_id(self):
        return self._ufl_id

    def ufl_domain(self):
        return Mesh(triangle, ufl_id=self.ufl_id(), cargo=self)

    def ufl_measure(
        self, integral_type="dx", subdomain_id="everywhere", metadata=None, subdomain_data=None
    ):
        return Measure(
            integral_type,
            subdomain_id=subdomain_id,
            metadata=metadata,
            domain=self,
            subdomain_data=subdomain_data,
        )


class MockMeshFunction:
    """Mock class for the pydolfin compatibility hack for domain data with [] syntax."""

    def __init__(self, ufl_id, mesh):
        self._mesh = mesh
        self._ufl_id = ufl_id

    def ufl_id(self):
        return self._ufl_id

    def mesh(self):
        return self._mesh

    def ufl_measure(self, integral_type=None, subdomain_id="everywhere", metadata=None):
        return Measure(
            integral_type,
            subdomain_id=subdomain_id,
            metadata=metadata,
            domain=self.mesh(),
            subdomain_data=self,
        )



from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dot,
    dx,
    grad,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = TestFunction(space)
u = TrialFunction(space)
u0 = Coefficient(space)
f = Coefficient(space)

a = (1 + u0**2) * dot(grad(v), grad(u)) * dx + 2 * u0 * u * dot(grad(v), grad(u0)) * dx
L = v * f * dx - (1 + u0**2) * dot(grad(v), grad(u0)) * dx



"""Extract part of a form in a mixed FunctionSpace."""

# Copyright (C) 2016-2024 Chris Richardson and Lawrence Mitchell
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Cecile Daversin-Catty, 2018

from typing import Optional

from ufl.algorithms.map_integrands import map_expr_dag, map_integrand_dags
from ufl.argument import Argument
from ufl.classes import FixedIndex, ListTensor
from ufl.constantvalue import Zero
from ufl.corealg.multifunction import MultiFunction
from ufl.functionspace import FunctionSpace
from ufl.tensors import as_vector


class FormSplitter(MultiFunction):
    """Form splitter."""

    def split(self, form, ix, iy=0):
        """Split."""
        # Remember which block to extract
        self.idx = [ix, iy]
        return map_integrand_dags(self, form)

    def argument(self, obj):
        """Apply to argument."""
        if obj.part() is not None:
            # Mixed element built from MixedFunctionSpace,
            # whose sub-function spaces are indexed by obj.part()
            if obj.part() == self.idx[obj.number()]:
                return obj
            else:
                return Zero(obj.ufl_shape)
        else:
            # Mixed element built from MixedElement,
            # whose sub-elements need their function space to be created
            Q = obj.ufl_function_space()
            dom = Q.ufl_domain()
            sub_elements = obj.ufl_element().sub_elements

            # If not a mixed element, do nothing
            if len(sub_elements) == 0:
                return obj

            args = []
            for i, sub_elem in enumerate(sub_elements):
                Q_i = FunctionSpace(dom, sub_elem)
                a = Argument(Q_i, obj.number(), part=obj.part())

                indices = [()]
                for m in a.ufl_shape:
                    indices = [(k + (j,)) for k in indices for j in range(m)]

                if i == self.idx[obj.number()]:
                    args += [a[j] for j in indices]
                else:
                    args += [Zero() for j in indices]

            return as_vector(args)

    def indexed(self, o, child, multiindex):
        """Extract indexed entry if multindices are fixed.

        This avoids tensors like (v_0, 0)[1] to be created.
        """
        indices = multiindex.indices()
        if isinstance(child, ListTensor) and all(isinstance(i, FixedIndex) for i in indices):
            if len(indices) == 1:
                return child.ufl_operands[indices[0]._value]
            else:
                return ListTensor(*(child.ufl_operands[i._value] for i in multiindex.indices()))
        return self.expr(o, child, multiindex)

    def multi_index(self, obj):
        """Apply to multi_index."""
        return obj

    def restricted(self, o):
        """Apply to a restricted function."""
        # If we hit a restriction first apply form splitter to argument, then check for zero
        op_split = map_expr_dag(self, o.ufl_operands[0])
        if isinstance(op_split, Zero):
            return op_split
        else:
            return op_split(o._side)

    expr = MultiFunction.reuse_if_untouched


def extract_blocks(form, i: Optional[int] = None, j: Optional[None] = None):
    """Extract blocks of a form.

    If arity is 0, returns the form.
    If arity is 1, return the ith block. If ``i`` is ``None``, return all blocks.
    If arity is 2, return the ``(i,j)`` entry. If ``j`` is ``None``, return the ith row.

    If neither `i` nor `j` are set, return all blocks (as a scalar, vector or tensor).

    Args:
        form: A form
        i: Index of the block to extract. If set to ``None``, ``j`` must be None.
        j: Index of the block to extract.
    """
    if i is None and j is not None:
        raise RuntimeError(f"Cannot extract block with {j=} and {i=}.")

    fs = FormSplitter()
    arguments = form.arguments()
    numbers = tuple(sorted(set(a.number() for a in arguments)))
    arity = len(numbers)
    assert arity <= 2
    if arity == 0:
        return (form,)

    # If mixed element, each argument has no sub-elements
    parts = tuple(sorted(set(part for a in arguments if (part := a.part()) is not None)))
    if parts == ():
        if i is None and j is None:
            num_sub_elements = arguments[0].ufl_element().num_sub_elements
            forms = []
            for pi in range(num_sub_elements):
                form_i = []
                for pj in range(num_sub_elements):
                    f = fs.split(form, pi, pj)
                    if f.empty():
                        form_i.append(None)
                    else:
                        form_i.append(f)
                forms.append(tuple(form_i))
            return tuple(forms)
        else:
            return fs.split(form, i, j)

    # If mixed function space, each argument has sub-elements
    forms = []
    num_parts = len(parts)
    for pi in range(num_parts):
        form_i = []
        if arity > 1:
            for pj in range(num_parts):
                f = fs.split(form, pi, pj)
                if f.empty():
                    form_i.append(None)
                else:
                    if (num_args := len(f.arguments())) != 2:
                        raise RuntimeError(f"Expected 2 arguments, got {num_args}")
                    form_i.append(f)
            forms.append(tuple(form_i))
        else:
            f = fs.split(form, pi)
            if f.empty():
                forms.append(None)
            else:
                forms.append(f)

    try:
        forms_tuple = tuple(forms)
    except TypeError:
        # Only one form returned
        forms_tuple = (forms,)
    if i is not None:
        if (num_rows := len(forms_tuple)) <= i:
            raise RuntimeError(f"Cannot extract block {i} from form with {num_rows} blocks.")
        if arity > 1 and j is not None:
            if (num_cols := len(forms_tuple[i])) <= j:
                raise RuntimeError(
                    f"Cannot extract block {i},{j} from form with {num_rows}x{num_cols} blocks."
                )
            return forms_tuple[i][j]
        else:
            return forms_tuple[i]
    else:
        return forms_tuple



__authors__ = "Martin Sandve Alnæs"
__date__ = "2009-02-13 -- 2009-02-13"

import math

from ufl import (
    Argument,
    Coefficient,
    FunctionSpace,
    Identity,
    Mesh,
    SpatialCoordinate,
    as_matrix,
    as_vector,
    cos,
    cross,
    det,
    dev,
    dot,
    exp,
    i,
    indices,
    inner,
    j,
    ln,
    outer,
    sin,
    skew,
    sqrt,
    sym,
    tan,
    tetrahedron,
    tr,
    triangle,
)
from ufl.constantvalue import as_ufl
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def testScalars():
    s = as_ufl(123)
    e = s((5, 7))
    v = 123
    assert e == v


def testZero():
    s = as_ufl(0)
    e = s((5, 7))
    v = 0
    assert e == v


def testIdentity():
    cell = triangle
    ident = Identity(cell.topological_dimension())

    s = 123 * ident[0, 0]
    e = s((5, 7))
    v = 123
    assert e == v

    s = 123 * ident[1, 0]
    e = s((5, 7))
    v = 0
    assert e == v


def testCoords():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    s = x[0] + x[1]
    e = s((5, 7))
    v = 5 + 7
    assert e == v


def testFunction1():
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    s = 3 * f
    e = s((5, 7), {f: 123})
    v = 3 * 123
    assert e == v


def testFunction2():
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)

    def g(x):
        return x[0]

    s = 3 * f
    e = s((5, 7), {f: g})
    v = 3 * 5
    assert e == v


def testArgument2():
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Argument(space, 2)

    def g(x):
        return x[0]

    s = 3 * f
    e = s((5, 7), {f: g})
    v = 3 * 5
    assert e == v


def testAlgebra():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    s = 3 * (x[0] + x[1]) - 7 + x[0] ** (x[1] / 2)
    e = s((5, 7))
    v = 3 * (5.0 + 7.0) - 7 + 5.0 ** (7.0 / 2)
    assert e == v


def testIndexSum():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    (i,) = indices(1)
    s = x[i] * x[i]
    e = s((5, 7))
    v = 5**2 + 7**2
    assert e == v


def testIndexSum2():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    ident = Identity(cell.topological_dimension())
    i, j = indices(2)
    s = (x[i] * x[j]) * ident[i, j]
    e = s((5, 7))
    # v = sum_i sum_j x_i x_j delta_ij = x_0 x_0 + x_1 x_1
    v = 5**2 + 7**2
    assert e == v


def testMathFunctions():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)[0]

    s = sin(x)
    e = s((5, 7))
    v = math.sin(5)
    assert e == v

    s = cos(x)
    e = s((5, 7))
    v = math.cos(5)
    assert e == v

    s = tan(x)
    e = s((5, 7))
    v = math.tan(5)
    assert e == v

    s = ln(x)
    e = s((5, 7))
    v = math.log(5)
    assert e == v

    s = exp(x)
    e = s((5, 7))
    v = math.exp(5)
    assert e == v

    s = sqrt(x)
    e = s((5, 7))
    v = math.sqrt(5)
    assert e == v


def testListTensor():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x, y = SpatialCoordinate(domain)

    m = as_matrix([[x, y], [-y, -x]])

    s = m[0, 0] + m[1, 0] + m[0, 1] + m[1, 1]
    e = s((5, 7))
    v = 0
    assert e == v

    s = m[0, 0] * m[1, 0] * m[0, 1] * m[1, 1]
    e = s((5, 7))
    v = 5**2 * 7**2
    assert e == v


def testComponentTensor1():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    m = as_vector(x[i], i)

    s = m[0] * m[1]
    e = s((5, 7))
    v = 5 * 7
    assert e == v


def testComponentTensor2():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xx = outer(x, x)

    m = as_matrix(xx[i, j], (i, j))

    s = m[0, 0] + m[1, 0] + m[0, 1] + m[1, 1]
    e = s((5, 7))
    v = 5 * 5 + 5 * 7 + 5 * 7 + 7 * 7
    assert e == v


def testComponentTensor3():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xx = outer(x, x)

    m = as_matrix(xx[i, j], (i, j))

    s = m[0, 0] * m[1, 0] * m[0, 1] * m[1, 1]
    e = s((5, 7))
    v = 5 * 5 * 5 * 7 * 5 * 7 * 7 * 7
    assert e == v


def testCoefficient():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    f = Coefficient(space)
    e = f**2

    def eval_f(x):
        return x[0] * x[1] ** 2

    assert e((3, 7), {f: eval_f}) == (3 * 7**2) ** 2


def testCoefficientDerivative():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    f = Coefficient(space)
    e = f.dx(0) ** 2 + f.dx(1) ** 2

    def eval_f(x, derivatives):
        if not derivatives:
            return eval_f.c * x[0] * x[1] ** 2
        # assume only first order derivative
        (d,) = derivatives
        if d == 0:
            return eval_f.c * x[1] ** 2
        if d == 1:
            return eval_f.c * x[0] * 2 * x[1]

    # shows how to attach data to eval_f
    eval_f.c = 5

    assert e((3, 7), {f: eval_f}) == (5 * 7**2) ** 2 + (5 * 3 * 2 * 7) ** 2


def test_dot():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    s = dot(x, 2 * x)
    e = s((5, 7))
    v = 2 * (5 * 5 + 7 * 7)
    assert e == v


def test_inner():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xx = as_matrix(((2 * x[0], 3 * x[0]), (2 * x[1], 3 * x[1])))
    s = inner(xx, 2 * xx)
    e = s((5, 7))
    v = 2 * ((5 * 2) ** 2 + (5 * 3) ** 2 + (7 * 2) ** 2 + (7 * 3) ** 2)
    assert e == v


def test_outer():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xx = outer(outer(x, x), as_vector((2, 3)))
    s = inner(xx, 2 * xx)
    e = s((5, 7))
    v = 2 * (5**2 + 7**2) ** 2 * (2**2 + 3**2)
    assert e == v


def test_cross():
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (3, 5, 7)

    # Test cross product of triplets of orthogonal
    # vectors, where |a x b| = |a| |b|
    ts = [
        [as_vector((x[0], 0, 0)), as_vector((0, x[1], 0)), as_vector((0, 0, x[2]))],
        [as_vector((x[0], x[1], 0)), as_vector((x[1], -x[0], 0)), as_vector((0, 0, x[2]))],
        [as_vector((0, x[0], x[1])), as_vector((0, x[1], -x[0])), as_vector((x[2], 0, 0))],
        [as_vector((x[0], 0, x[1])), as_vector((x[1], 0, -x[0])), as_vector((0, x[2], 0))],
    ]
    for t in ts:
        for a in range(3):
            for b in range(3):
                cab = cross(t[a], t[b])
                dab = dot(cab, cab)
                eab = dab(xv)
                tna = dot(t[a], t[a])(xv)
                tnb = dot(t[b], t[b])(xv)
                vab = tna * tnb if a != b else 0
                assert eab == vab


def xtest_dev():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (5, 7)
    xx = outer(x, x)
    s1 = dev(2 * xx)
    s2 = 2 * (xx - xx.T)  # FIXME
    e = inner(s1, s1)(xv)
    v = inner(s2, s2)(xv)
    assert e == v


def test_skew():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (5, 7)
    xx = outer(x, x)
    s1 = skew(2 * xx)
    s2 = xx - xx.T
    e = inner(s1, s1)(xv)
    v = inner(s2, s2)(xv)
    assert e == v


def test_sym():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (5, 7)
    xx = outer(x, x)
    s1 = sym(2 * xx)
    s2 = xx + xx.T
    e = inner(s1, s1)(xv)
    v = inner(s2, s2)(xv)
    assert e == v


def test_tr():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (5, 7)
    xx = outer(x, x)
    s = tr(2 * xx)
    e = s(xv)
    v = 2 * sum(xv[i] ** 2 for i in (0, 1))
    assert e == v


def test_det2D():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    xv = (5, 7)
    a, b = 6.5, -4
    xx = as_matrix(((x[0], x[1]), (a, b)))
    s = det(2 * xx)
    e = s(xv)
    v = 2**2 * (5 * b - 7 * a)
    assert e == v


def xtest_det3D():  # FIXME
    x = SpatialCoordinate(tetrahedron)
    xv = (5, 7, 9)
    a, b, c = 6.5, -4, 3
    d, e, f = 2, 3, 4
    xx = as_matrix(((x[0], x[1], x[2]), (a, b, c), (d, e, f)))
    s = det(2 * xx)
    e = s(xv)
    v = 2**3 * (xv[0] * (b * f - e * c) - xv[1] * (a * f - c * d) + xv[2] * (a * e - b * d))
    assert e == v


def test_cofac():
    pass  # TODO


def test_inv():
    pass  # TODO



"""Apply derivatives algorithm which computes the derivatives of a form of expression."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import warnings
from collections import defaultdict
from math import pi

from ufl.action import Action
from ufl.algorithms.analysis import extract_arguments
from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.algorithms.replace_derivative_nodes import replace_derivative_nodes
from ufl.argument import BaseArgument
from ufl.checks import is_cellwise_constant
from ufl.classes import (
    Coefficient,
    ComponentTensor,
    Conj,
    ConstantValue,
    ExprList,
    ExprMapping,
    FloatValue,
    FormArgument,
    Grad,
    Identity,
    Imag,
    Indexed,
    IndexSum,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    ListTensor,
    Product,
    Real,
    ReferenceGrad,
    ReferenceValue,
    SpatialCoordinate,
    Sum,
    Variable,
    Zero,
)
from ufl.constantvalue import is_true_ufl_scalar, is_ufl_scalar
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.expr import ufl_err_str
from ufl.core.multiindex import FixedIndex, MultiIndex, indices
from ufl.core.terminal import Terminal
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction
from ufl.differentiation import (
    BaseFormCoordinateDerivative,
    BaseFormOperatorDerivative,
    CoordinateDerivative,
)
from ufl.domain import extract_unique_domain
from ufl.form import Form, ZeroBaseForm
from ufl.operators import (
    bessel_I,
    bessel_J,
    bessel_K,
    bessel_Y,
    cell_avg,
    conditional,
    cos,
    cosh,
    exp,
    facet_avg,
    ln,
    sign,
    sin,
    sinh,
    sqrt,
)
from ufl.pullback import CustomPullback, PhysicalPullback
from ufl.tensors import as_scalar, as_scalars, as_tensor, unit_indexed_tensor, unwrap_list_tensor

# TODO: Add more rulesets?
# - DivRuleset
# - CurlRuleset
# - ReferenceGradRuleset
# - ReferenceDivRuleset


class GenericDerivativeRuleset(MultiFunction):
    """A generic derivative."""

    def __init__(self, var_shape):
        """Initialise."""
        MultiFunction.__init__(self)
        self._var_shape = var_shape

    # --- Error checking for missing handlers and unexpected types

    def expr(self, o):
        """Raise error."""
        raise ValueError(
            f"Missing differentiation handler for type {o._ufl_class_.__name__}. "
            "Have you added a new type?"
        )

    def unexpected(self, o):
        """Raise error about unexpected type."""
        raise ValueError(f"Unexpected type {o._ufl_class_.__name__} in AD rules.")

    def override(self, o):
        """Raise error about overriding."""
        raise ValueError(
            f"Type {o._ufl_class_.__name__} must be overridden in specialized AD rule set."
        )

    def derivative(self, o):
        """Raise error."""
        raise ValueError(
            f"Unhandled derivative type {o._ufl_class_.__name__}, "
            "nested differentiation has failed."
        )

    # --- Some types just don't have any derivative, this is just to
    # --- make algorithm structure generic

    def non_differentiable_terminal(self, o):
        """Return the non-differentiated object.

        Labels and indices are not differentiable: it's convenient to
        return the non-differentiated object.
        """
        return o

    label = non_differentiable_terminal
    multi_index = non_differentiable_terminal

    # --- Helper functions for creating zeros with the right shapes

    def independent_terminal(self, o):
        """A zero with correct shape for terminals independent of diff. variable."""
        return Zero(o.ufl_shape + self._var_shape)

    def independent_operator(self, o):
        """A zero with correct shape and indices for operators independent of diff. variable."""
        return Zero(o.ufl_shape + self._var_shape, o.ufl_free_indices, o.ufl_index_dimensions)

    # --- All derivatives need to define grad and averaging

    grad = override
    cell_avg = override
    facet_avg = override

    # --- Default rules for terminals

    # Literals are by definition independent of any differentiation variable
    constant_value = independent_terminal

    # Constants are independent of any differentiation
    constant = independent_terminal

    # Rules for form arguments must be specified in specialized rule set
    form_argument = override

    # Rules for geometric quantities must be specified in specialized rule set
    geometric_quantity = override

    # These types are currently assumed independent, but for non-affine domains
    # this no longer holds and we want to implement rules for them.
    # facet_normal = independent_terminal
    # spatial_coordinate = independent_terminal
    # cell_coordinate = independent_terminal

    # Measures of cell entities, assuming independent although
    # this will not be true for all of these for non-affine domains
    # cell_volume = independent_terminal
    # circumradius = independent_terminal
    # facet_area = independent_terminal
    # cell_surface_area = independent_terminal
    # min_cell_edge_length = independent_terminal
    # max_cell_edge_length = independent_terminal
    # min_facet_edge_length = independent_terminal
    # max_facet_edge_length = independent_terminal

    # Other stuff
    # cell_orientation = independent_terminal
    # quadrature_weigth = independent_terminal

    # These types are currently not expected to show up in AD pass.
    # To make some of these available to the end-user, they need to be
    # implemented here.
    # facet_coordinate = unexpected
    # cell_origin = unexpected
    # facet_origin = unexpected
    # cell_facet_origin = unexpected
    # jacobian = unexpected
    # jacobian_determinant = unexpected
    # jacobian_inverse = unexpected
    # facet_jacobian = unexpected
    # facet_jacobian_determinant = unexpected
    # facet_jacobian_inverse = unexpected
    # cell_facet_jacobian = unexpected
    # cell_facet_jacobian_determinant = unexpected
    # cell_facet_jacobian_inverse = unexpected
    # cell_vertices = unexpected
    # cell_edge_vectors = unexpected
    # facet_edge_vectors = unexpected
    # reference_cell_edge_vectors = unexpected
    # reference_facet_edge_vectors = unexpected
    # cell_normal = unexpected # TODO: Expecting rename
    # cell_normals = unexpected
    # facet_tangents = unexpected
    # cell_tangents = unexpected
    # cell_midpoint = unexpected
    # facet_midpoint = unexpected

    # --- Default rules for operators

    def variable(self, o, df, unused_l):
        """Differentiate a variable."""
        return df

    # --- Indexing and component handling

    def indexed(self, o, Ap, ii):  # TODO: (Partially) duplicated in nesting rules
        """Differentiate an indexed."""
        # Propagate zeros
        if isinstance(Ap, Zero):
            return self.independent_operator(o)

        # Untangle as_tensor(C[kk], jj)[ii] -> C[ll] to simplify
        # resulting expression
        if isinstance(Ap, ComponentTensor):
            B, jj = Ap.ufl_operands
            if isinstance(B, Indexed):
                C, kk = B.ufl_operands
                kk = list(kk)
                if all(j in kk for j in jj):
                    rep = dict(zip(jj, ii))
                    Cind = [rep.get(k, k) for k in kk]
                    expr = Indexed(C, MultiIndex(tuple(Cind)))
                    assert expr.ufl_free_indices == o.ufl_free_indices
                    assert expr.ufl_shape == o.ufl_shape
                    return expr

        # Otherwise a more generic approach
        r = len(Ap.ufl_shape) - len(ii)
        if r:
            kk = indices(r)
            op = Indexed(Ap, MultiIndex(ii.indices() + kk))
            op = as_tensor(op, kk)
        else:
            op = Indexed(Ap, ii)
        return op

    def list_tensor(self, o, *dops):
        """Differentiate a list_tensor."""
        return ListTensor(*dops)

    def component_tensor(self, o, Ap, ii):
        """Differentiate a component_tensor."""
        if isinstance(Ap, Zero):
            op = self.independent_operator(o)
        else:
            Ap, jj = as_scalar(Ap)
            op = as_tensor(Ap, ii.indices() + jj)
        return op

    # --- Algebra operators

    def index_sum(self, o, Ap, i):
        """Differentiate an index_sum."""
        return IndexSum(Ap, i)

    def sum(self, o, da, db):
        """Differentiate a sum."""
        return da + db

    def product(self, o, da, db):
        """Differentiate a product."""
        # Even though arguments to o are scalar, da and db may be
        # tensor valued
        a, b = o.ufl_operands
        (da, db), ii = as_scalars(da, db)
        pa = Product(da, b)
        pb = Product(a, db)
        s = Sum(pa, pb)
        if ii:
            s = as_tensor(s, ii)
        return s

    def division(self, o, fp, gp):
        """Differentiate a division."""
        f, g = o.ufl_operands

        if not is_ufl_scalar(f):
            raise ValueError("Not expecting nonscalar nominator")
        if not is_true_ufl_scalar(g):
            raise ValueError("Not expecting nonscalar denominator")

        # do_df = 1/g
        # do_dg = -h/g
        # op = do_df*fp + do_df*gp
        # op = (fp - o*gp) / g

        # Get o and gp as scalars, multiply, then wrap as a tensor
        # again
        so, oi = as_scalar(o)
        sgp, gi = as_scalar(gp)
        o_gp = so * sgp
        if oi or gi:
            o_gp = as_tensor(o_gp, oi + gi)
        op = (fp - o_gp) / g

        return op

    def power(self, o, fp, gp):
        """Differentiate a power."""
        f, g = o.ufl_operands

        if not is_true_ufl_scalar(f):
            raise ValueError("Expecting scalar expression f in f**g.")
        if not is_true_ufl_scalar(g):
            raise ValueError("Expecting scalar expression g in f**g.")

        # Derivation of the general case: o = f(x)**g(x)
        # do/df  = g * f**(g-1) = g / f * o
        # do/dg  = ln(f) * f**g = ln(f) * o
        # do/df * df + do/dg * dg = o * (g / f * df + ln(f) * dg)

        if isinstance(gp, Zero):
            # This probably produces better results for the common
            # case of f**constant
            op = fp * g * f ** (g - 1)
        else:
            # Note: This produces expressions like (1/w)*w**5 instead of w**4
            # op = o * (fp * g / f + gp * ln(f)) # This reuses o
            op = f ** (g - 1) * (
                g * fp + f * ln(f) * gp
            )  # This gives better accuracy in dolfin integration test

        # Example: d/dx[x**(x**3)]:
        # f = x
        # g = x**3
        # df = 1
        # dg = 3*x**2
        # op1 = o * (fp * g / f + gp * ln(f))
        #     = x**(x**3)   * (x**3/x + 3*x**2*ln(x))
        # op2 = f**(g-1) * (g*fp + f*ln(f)*gp)
        #     = x**(x**3-1) * (x**3 + x*3*x**2*ln(x))

        return op

    def abs(self, o, df):
        """Differentiate an abs."""
        (f,) = o.ufl_operands
        # return conditional(eq(f, 0), 0, Product(sign(f), df)) abs is
        # not complex differentiable, so we workaround the case of a
        # real F in complex mode by defensively casting to real inside
        # the sign.
        return sign(Real(f)) * df

    # --- Complex algebra

    def conj(self, o, df):
        """Differentiate a conj."""
        return Conj(df)

    def real(self, o, df):
        """Differentiate a real."""
        return Real(df)

    def imag(self, o, df):
        """Differentiate a imag."""
        return Imag(df)

    # --- Mathfunctions

    def math_function(self, o, df):
        """Differentiate a math_function."""
        # FIXME: Introduce a UserOperator type instead of this hack
        # and define user derivative() function properly
        if hasattr(o, "derivative"):
            (f,) = o.ufl_operands
            return df * o.derivative()
        raise ValueError("Unknown math function.")

    def sqrt(self, o, fp):
        """Differentiate a sqrt."""
        return fp / (2 * o)

    def exp(self, o, fp):
        """Differentiate an exp."""
        return fp * o

    def ln(self, o, fp):
        """Differentiate a ln."""
        (f,) = o.ufl_operands
        if isinstance(f, Zero):
            raise ZeroDivisionError()
        return fp / f

    def cos(self, o, fp):
        """Differentiate a cos."""
        (f,) = o.ufl_operands
        return fp * -sin(f)

    def sin(self, o, fp):
        """Differentiate a sin."""
        (f,) = o.ufl_operands
        return fp * cos(f)

    def tan(self, o, fp):
        """Differentiate a tan."""
        (f,) = o.ufl_operands
        return 2.0 * fp / (cos(2.0 * f) + 1.0)

    def cosh(self, o, fp):
        """Differentiate a cosh."""
        (f,) = o.ufl_operands
        return fp * sinh(f)

    def sinh(self, o, fp):
        """Differentiate a sinh."""
        (f,) = o.ufl_operands
        return fp * cosh(f)

    def tanh(self, o, fp):
        """Differentiate a tanh."""
        (f,) = o.ufl_operands

        def sech(y):
            return (2.0 * cosh(y)) / (cosh(2.0 * y) + 1.0)

        return fp * sech(f) ** 2

    def acos(self, o, fp):
        """Differentiate an acos."""
        (f,) = o.ufl_operands
        return -fp / sqrt(1.0 - f**2)

    def asin(self, o, fp):
        """Differentiate an asin."""
        (f,) = o.ufl_operands
        return fp / sqrt(1.0 - f**2)

    def atan(self, o, fp):
        """Differentiate an atan."""
        (f,) = o.ufl_operands
        return fp / (1.0 + f**2)

    def atan2(self, o, fp, gp):
        """Differentiate an atan2."""
        f, g = o.ufl_operands
        return (g * fp - f * gp) / (f**2 + g**2)

    def erf(self, o, fp):
        """Differentiate an erf."""
        (f,) = o.ufl_operands
        return fp * (2.0 / sqrt(pi) * exp(-(f**2)))

    # --- Bessel functions

    def bessel_j(self, o, nup, fp):
        """Differentiate a bessel_j."""
        nu, f = o.ufl_operands
        if not (nup is None or isinstance(nup, Zero)):
            raise NotImplementedError(
                "Differentiation of bessel function w.r.t. nu is not supported."
            )

        if isinstance(nu, Zero):
            op = -bessel_J(1, f)
        else:
            op = 0.5 * (bessel_J(nu - 1, f) - bessel_J(nu + 1, f))
        return op * fp

    def bessel_y(self, o, nup, fp):
        """Differentiate a bessel_y."""
        nu, f = o.ufl_operands
        if not (nup is None or isinstance(nup, Zero)):
            raise NotImplementedError(
                "Differentiation of bessel function w.r.t. nu is not supported."
            )

        if isinstance(nu, Zero):
            op = -bessel_Y(1, f)
        else:
            op = 0.5 * (bessel_Y(nu - 1, f) - bessel_Y(nu + 1, f))
        return op * fp

    def bessel_i(self, o, nup, fp):
        """Differentiate a bessel_i."""
        nu, f = o.ufl_operands
        if not (nup is None or isinstance(nup, Zero)):
            raise NotImplementedError(
                "Differentiation of bessel function w.r.t. nu is not supported."
            )

        if isinstance(nu, Zero):
            op = bessel_I(1, f)
        else:
            op = 0.5 * (bessel_I(nu - 1, f) + bessel_I(nu + 1, f))
        return op * fp

    def bessel_k(self, o, nup, fp):
        """Differentiate a bessel_k."""
        nu, f = o.ufl_operands
        if not (nup is None or isinstance(nup, Zero)):
            raise NotImplementedError(
                "Differentiation of bessel function w.r.t. nu is not supported."
            )

        if isinstance(nu, Zero):
            op = -bessel_K(1, f)
        else:
            op = -0.5 * (bessel_K(nu - 1, f) + bessel_K(nu + 1, f))
        return op * fp

    # --- Restrictions

    def restricted(self, o, fp):
        """Differentiate a restricted."""
        # Restriction and differentiation commutes
        if isinstance(fp, ConstantValue):
            return fp  # TODO: Add simplification to Restricted instead?
        else:
            return fp(o._side)  # (f+-)' == (f')+-

    # --- Conditionals

    def binary_condition(self, o, dl, dr):
        """Differentiate a binary_condition."""
        # Should not be used anywhere...
        return None

    def not_condition(self, o, c):
        """Differentiate a not_condition."""
        # Should not be used anywhere...
        return None

    def conditional(self, o, unused_dc, dt, df):
        """Differentiate a conditional."""
        if isinstance(dt, Zero) and isinstance(df, Zero):
            # Assuming dt and df have the same indices here, which
            # should be the case
            return dt
        else:
            # Not placing t[1],f[1] outside, allowing arguments inside
            # conditionals.  This will make legacy ffc fail, but
            # should work with uflacs.
            c = o.ufl_operands[0]
            return conditional(c, dt, df)

    def max_value(self, o, df, dg):
        """Differentiate a max_value."""
        # d/dx max(f, g) =
        # f > g: df/dx
        # f < g: dg/dx
        # Placing df,dg outside here to avoid getting arguments inside
        # conditionals
        f, g = o.ufl_operands
        dc = conditional(f > g, 1, 0)
        return dc * df + (1.0 - dc) * dg

    def min_value(self, o, df, dg):
        """Differentiate a min_value."""
        # d/dx min(f, g) =
        #  f < g: df/dx
        #  else: dg/dx
        #  Placing df,dg outside here to avoid getting arguments
        #  inside conditionals
        f, g = o.ufl_operands
        dc = conditional(f < g, 1, 0)
        return dc * df + (1.0 - dc) * dg


class GradRuleset(GenericDerivativeRuleset):
    """Take the grad derivative."""

    def __init__(self, geometric_dimension):
        """Initialise."""
        GenericDerivativeRuleset.__init__(self, var_shape=(geometric_dimension,))
        self._Id = Identity(geometric_dimension)

    # --- Specialized rules for geometric quantities

    def geometric_quantity(self, o):
        """Differentiate a geometric_quantity.

        Default for geometric quantities is do/dx = 0 if piecewise constant,
        otherwise transform derivatives to reference derivatives.
        Override for specific types if other behaviour is needed.
        """
        if is_cellwise_constant(o):
            return self.independent_terminal(o)
        else:
            domain = extract_unique_domain(o)
            K = JacobianInverse(domain)
            Do = grad_to_reference_grad(o, K)
            return Do

    def jacobian_inverse(self, o):
        """Differentiate a jacobian_inverse."""
        # grad(K) == K_ji rgrad(K)_rj
        if is_cellwise_constant(o):
            return self.independent_terminal(o)
        if not o._ufl_is_terminal_:
            raise ValueError("ReferenceValue can only wrap a terminal")
        Do = grad_to_reference_grad(o, o)
        return Do

    # TODO: Add more explicit geometry type handlers here, with
    # non-affine domains several should be non-zero.

    def spatial_coordinate(self, o):
        """Differentiate a spatial_coordinate.

        dx/dx = I.
        """
        return self._Id

    def cell_coordinate(self, o):
        """Differentiate a cell_coordinate.

        dX/dx = inv(dx/dX) = inv(J) = K.
        """
        # FIXME: Is this true for manifolds? What about orientation?
        return JacobianInverse(extract_unique_domain(o))

    # --- Specialized rules for form arguments

    def base_form_operator(self, o):
        """Differentiate a base_form_operator."""
        # Push the grad through the operator is not legal in most cases:
        #    -> Not enouth regularity for chain rule to hold!
        # By the time we evaluate `grad(o)`, the operator `o` will have
        # been assembled and substituted by its output.
        return Grad(o)

    def coefficient(self, o):
        """Differentiate a coefficient."""
        if is_cellwise_constant(o):
            return self.independent_terminal(o)
        return Grad(o)

    def argument(self, o):
        """Differentiate an argument."""
        # TODO: Enable this after fixing issue#13, unless we move
        # simplificat ion to a separate stage?
        # if is_cellwise_constant(o):
        #     # Collapse gradient of cellwise constant function to zero
        #     # TODO: Missing this type
        #     return AnnotatedZero(o.ufl_shape + self._var_shape, arguments=(o,))
        return Grad(o)

    # --- Rules for values or derivatives in reference frame

    def reference_value(self, o):
        """Differentiate a reference_value."""
        # grad(o) == grad(rv(f)) -> K_ji*rgrad(rv(f))_rj
        f = o.ufl_operands[0]
        if isinstance(f.ufl_element().pullback, PhysicalPullback):
            # TODO: Do we need to be more careful for immersed things?
            return ReferenceGrad(o)

        if not f._ufl_is_terminal_:
            raise ValueError("ReferenceValue can only wrap a terminal")
        domain = extract_unique_domain(f)
        K = JacobianInverse(domain)
        Do = grad_to_reference_grad(o, K)
        return Do

    def reference_grad(self, o):
        """Differentiate a reference_grad."""
        # grad(o) == grad(rgrad(rv(f))) -> K_ji*rgrad(rgrad(rv(f)))_rj
        f = o.ufl_operands[0]

        valid_operand = f._ufl_is_in_reference_frame_ or isinstance(
            f, (JacobianInverse, SpatialCoordinate, Jacobian, JacobianDeterminant)
        )
        if not valid_operand:
            raise ValueError("ReferenceGrad can only wrap a reference frame type!")
        domain = extract_unique_domain(f)
        K = JacobianInverse(domain)
        Do = grad_to_reference_grad(o, K)
        return Do

    # --- Nesting of gradients

    def grad(self, o):
        """Differentiate a grad.

        Represent grad(grad(f)) as Grad(Grad(f)).
        """
        # Check that o is a "differential terminal"
        if not isinstance(o.ufl_operands[0], (Grad, Terminal)):
            raise ValueError("Expecting only grads applied to a terminal.")

        return Grad(o)

    def _grad(self, o):
        """Differentiate a _grad."""
        pass
        # TODO: Not sure how to detect that gradient of f is cellwise constant.
        #       Can we trust element degrees?
        # if is_cellwise_constant(o):
        #     return self.terminal(o)
        # TODO: Maybe we can ask "f.has_derivatives_of_order(n)" to check
        #       if we should make a zero here?
        # 1) n = count number of Grads, get f
        # 2) if not f.has_derivatives(n): return zero(...)

    cell_avg = GenericDerivativeRuleset.independent_operator
    facet_avg = GenericDerivativeRuleset.independent_operator


def grad_to_reference_grad(o, K):
    """Relates grad(o) to reference_grad(o) using the Jacobian inverse.

    Args:
        o: Operand
        K: Jacobian inverse
    Returns:
        grad(o) written in terms of reference_grad(o) and K
    """
    r = indices(len(o.ufl_shape))
    i, j = indices(2)
    # grad(o) == K_ji rgrad(o)_rj
    Do = as_tensor(K[j, i] * ReferenceGrad(o)[r + (j,)], r + (i,))
    return Do


class ReferenceGradRuleset(GenericDerivativeRuleset):
    """Apply the reference grad derivative."""

    def __init__(self, topological_dimension):
        """Initialise."""
        GenericDerivativeRuleset.__init__(self, var_shape=(topological_dimension,))
        self._Id = Identity(topological_dimension)

    # --- Specialized rules for geometric quantities

    def geometric_quantity(self, o):
        """Differentiate a geometric_quantity.

        dg/dX = 0 if piecewise constant, otherwise ReferenceGrad(g).
        """
        if is_cellwise_constant(o):
            return self.independent_terminal(o)
        else:
            # TODO: Which types does this involve? I don't think the
            # form compilers will handle this.
            return ReferenceGrad(o)

    def spatial_coordinate(self, o):
        """Differentiate a spatial_coordinate.

        dx/dX = J.
        """
        # Don't convert back to J, otherwise we get in a loop
        return ReferenceGrad(o)

    def cell_coordinate(self, o):
        """Differentiate a cell_coordinate.

        dX/dX = I.
        """
        return self._Id

    # TODO: Add more geometry types here, with non-affine domains
    # several should be non-zero.

    # --- Specialized rules for form arguments

    def reference_value(self, o):
        """Differentiate a reference_value."""
        if not o.ufl_operands[0]._ufl_is_terminal_:
            raise ValueError("ReferenceValue can only wrap a terminal")
        return ReferenceGrad(o)

    def coefficient(self, o):
        """Differentiate a coefficient."""
        raise ValueError("Coefficient should be wrapped in ReferenceValue by now")

    def argument(self, o):
        """Differentiate an argument."""
        raise ValueError("Argument should be wrapped in ReferenceValue by now")

    # --- Nesting of gradients

    def grad(self, o):
        """Differentiate a grad."""
        raise ValueError(
            f"Grad should have been transformed by this point, but got {type(o).__name__}."
        )

    def reference_grad(self, o):
        """Differentiate a reference_grad.

        Represent ref_grad(ref_grad(f)) as RefGrad(RefGrad(f)).
        """
        # Check that o is a "differential terminal"
        if not isinstance(o.ufl_operands[0], (ReferenceGrad, ReferenceValue, Terminal)):
            raise ValueError("Expecting only grads applied to a terminal.")
        return ReferenceGrad(o)

    cell_avg = GenericDerivativeRuleset.independent_operator
    facet_avg = GenericDerivativeRuleset.independent_operator


class VariableRuleset(GenericDerivativeRuleset):
    """Differentiate with respect to a variable."""

    def __init__(self, var):
        """Initialise."""
        GenericDerivativeRuleset.__init__(self, var_shape=var.ufl_shape)
        if var.ufl_free_indices:
            raise ValueError("Differentiation variable cannot have free indices.")
        self._variable = var
        self._Id = self._make_identity(self._var_shape)

    def _make_identity(self, sh):
        """Differentiate a _make_identity.

        Creates a higher order identity tensor to represent dv/dv.
        """
        res = None
        if sh == ():
            # Scalar dv/dv is scalar
            return FloatValue(1.0)
        elif len(sh) == 1:
            # Vector v makes dv/dv the identity matrix
            return Identity(sh[0])
        else:
            # TODO: Add a type for this higher order identity?
            # II[i0,i1,i2,j0,j1,j2] = 1 if all((i0==j0, i1==j1, i2==j2)) else 0
            # Tensor v makes dv/dv some kind of higher rank identity tensor
            ind1 = ()
            ind2 = ()
            for d in sh:
                i, j = indices(2)
                dij = Identity(d)[i, j]
                if res is None:
                    res = dij
                else:
                    res *= dij
                ind1 += (i,)
                ind2 += (j,)
            fp = as_tensor(res, ind1 + ind2)
        return fp

    # Explicitly defining dg/dw == 0
    geometric_quantity = GenericDerivativeRuleset.independent_terminal

    # Explicitly defining da/dw == 0
    argument = GenericDerivativeRuleset.independent_terminal

    def coefficient(self, o):
        """Differentiate a coefficient.

        df/dv = Id if v is f else 0.

        Note that if v = variable(f), df/dv is still 0,
        but if v == f, i.e. isinstance(v, Coefficient) == True,
        then df/dv == df/df = Id.
        """
        v = self._variable
        if isinstance(v, Coefficient) and o == v:
            # dv/dv = identity of rank 2*rank(v)
            return self._Id
        else:
            # df/v = 0
            return self.independent_terminal(o)

    def variable(self, o, df, a):
        """Differentiate a variable."""
        v = self._variable
        if isinstance(v, Variable) and v.label() == a:
            # dv/dv = identity of rank 2*rank(v)
            return self._Id
        else:
            # df/v = df
            return df

    def grad(self, o):
        """Differentiate a grad.

        Variable derivative of a gradient of a terminal must be 0.
        """
        # Check that o is a "differential terminal"
        if not isinstance(o.ufl_operands[0], (Grad, Terminal)):
            raise ValueError("Expecting only grads applied to a terminal.")
        return self.independent_terminal(o)

    # --- Rules for values or derivatives in reference frame

    def reference_value(self, o):
        """Differentiate a reference_value."""
        # d/dv(o) == d/dv(rv(f)) = 0 if v is not f, or rv(dv/df)
        v = self._variable
        if isinstance(v, Coefficient) and o.ufl_operands[0] == v:
            if not v.ufl_element().pullback.is_identity:
                # FIXME: This is a bit tricky, instead of Identity it is
                #   actually inverse(transform), or we should rather not
                #   convert to reference frame in the first place
                raise ValueError(
                    "Missing implementation: To handle derivatives of rv(f) w.r.t. f for "
                    "mapped elements, rewriting to reference frame should not happen first..."
                )
            # dv/dv = identity of rank 2*rank(v)
            return self._Id
        else:
            # df/v = 0
            return self.independent_terminal(o)

    def reference_grad(self, o):
        """Differentiate a reference_grad.

        Variable derivative of a gradient of a terminal must be 0.
        """
        if not isinstance(o.ufl_operands[0], (ReferenceGrad, ReferenceValue)):
            raise ValueError("Unexpected argument to reference_grad.")
        return self.independent_terminal(o)

    cell_avg = GenericDerivativeRuleset.independent_operator
    facet_avg = GenericDerivativeRuleset.independent_operator


class GateauxDerivativeRuleset(GenericDerivativeRuleset):
    """Apply AFD (Automatic Functional Differentiation) to expression.

    Implements rules for the Gateaux derivative D_w[v](...) defined as
    D_w[v](e) = d/dtau e(w+tau v)|tau=0.
    """

    def __init__(self, coefficients, arguments, coefficient_derivatives, pending_operations):
        """Initialise."""
        GenericDerivativeRuleset.__init__(self, var_shape=())

        # Type checking
        if not isinstance(coefficients, ExprList):
            raise ValueError("Expecting a ExprList of coefficients.")
        if not isinstance(arguments, ExprList):
            raise ValueError("Expecting a ExprList of arguments.")
        if not isinstance(coefficient_derivatives, ExprMapping):
            raise ValueError("Expecting a coefficient-coefficient ExprMapping.")

        # The coefficient(s) to differentiate w.r.t. and the
        # argument(s) s.t. D_w[v](e) = d/dtau e(w+tau v)|tau=0
        self._w = coefficients.ufl_operands
        self._v = arguments.ufl_operands
        self._w2v = {w: v for w, v in zip(self._w, self._v)}

        # Build more convenient dict {f: df/dw} for each coefficient f
        # where df/dw is nonzero
        cd = coefficient_derivatives.ufl_operands
        self._cd = {cd[2 * i]: cd[2 * i + 1] for i in range(len(cd) // 2)}

        # Record the operations delayed to the derivative expansion phase:
        # Example: dN(u)/du where `N` is an ExternalOperator and `u` a Coefficient
        self.pending_operations = pending_operations

    # Explicitly defining dg/dw == 0
    geometric_quantity = GenericDerivativeRuleset.independent_terminal

    def cell_avg(self, o, fp):
        """Differentiate a cell_avg."""
        # Cell average of a single function and differentiation
        # commutes, D_f[v](cell_avg(f)) = cell_avg(v)
        return cell_avg(fp)

    def facet_avg(self, o, fp):
        """Differentiate a facet_avg."""
        # Facet average of a single function and differentiation
        # commutes, D_f[v](facet_avg(f)) = facet_avg(v)
        return facet_avg(fp)

    # Explicitly defining da/dw == 0
    argument = GenericDerivativeRuleset.independent_terminal

    def coefficient(self, o):
        """Differentiate a coefficient."""
        # Define dw/dw := d/ds [w + s v] = v

        # Return corresponding argument if we can find o among w
        do = self._w2v.get(o)
        if do is not None:
            return do

        # Look for o among coefficient derivatives
        dos = self._cd.get(o)
        if dos is None:
            # If o is not among coefficient derivatives, return
            # do/dw=0
            do = Zero(o.ufl_shape)
            return do
        else:
            # Compute do/dw_j = do/dw_h : v.
            # Since we may actually have a tuple of oprimes and vs in a
            # 'mixed' space, sum over them all to get the complete inner
            # product. Using indices to define a non-compound inner product.

            # Example:
            # (f:g) -> (dfdu:v):g + f:(dgdu:v)
            # shape(dfdu) == shape(f) + shape(v)
            # shape(f) == shape(g) == shape(dfdu : v)

            # Make sure we have a tuple to match the self._v tuple
            if not isinstance(dos, tuple):
                dos = (dos,)
            if len(dos) != len(self._v):
                raise ValueError(
                    "Got a tuple of arguments, expecting a "
                    "matching tuple of coefficient derivatives."
                )
            dosum = Zero(o.ufl_shape)
            for do, v in zip(dos, self._v):
                so, oi = as_scalar(do)
                rv = len(oi) - len(v.ufl_shape)
                oi1 = oi[:rv]
                oi2 = oi[rv:]
                prod = so * v[oi2]
                if oi1:
                    dosum += as_tensor(prod, oi1)
                else:
                    dosum += prod
            return dosum

    def reference_value(self, o):
        """Differentiate a reference_value."""
        raise NotImplementedError(
            "Currently no support for ReferenceValue in CoefficientDerivative."
        )
        # TODO: This is implementable for regular derivative(M(f),f,v)
        #       but too messy if customized coefficient derivative
        #       relations are given by the user.  We would only need
        #       this to allow the user to write
        #       derivative(...ReferenceValue...,...).
        # f, = o.ufl_operands
        # if not f._ufl_is_terminal_:
        #     raise ValueError("ReferenceValue can only wrap terminals directly.")
        # FIXME: check all cases like in coefficient
        # if f is w:
        #     # FIXME: requires that v is an Argument with the same element mapping!
        #     return ReferenceValue(v)
        # else:
        #     return self.independent_terminal(o)

    def reference_grad(self, o):
        """Differentiate a reference_grad."""
        raise NotImplementedError(
            "Currently no support for ReferenceGrad in CoefficientDerivative."
        )
        # TODO: This is implementable for regular derivative(M(f),f,v)
        #       but too messy if customized coefficient derivative
        #       relations are given by the user.  We would only need
        #       this to allow the user to write
        #       derivative(...ReferenceValue...,...).

    def grad(self, g):
        """Differentiate a grad."""
        # If we hit this type, it has already been propagated to a
        # coefficient (or grad of a coefficient) or a base form operator, # FIXME: Assert
        # this!  so we need to take the gradient of the variation or
        # return zero.  Complications occur when dealing with
        # derivatives w.r.t. single components...

        # Figure out how many gradients are around the inner terminal
        ngrads = 0
        o = g
        while isinstance(o, Grad):
            (o,) = o.ufl_operands
            ngrads += 1
        # `grad(N)` where N is a BaseFormOperator is treated as if `N` was a Coefficient.
        if not isinstance(o, (FormArgument, BaseFormOperator)):
            raise ValueError(f"Expecting gradient of a FormArgument, not {ufl_err_str(o)}.")

        def apply_grads(f):
            for i in range(ngrads):
                f = Grad(f)
            return f

        # Find o among all w without any indexing, which makes this
        # easy
        for w, v in zip(self._w, self._v):
            if o == w and isinstance(v, FormArgument):
                # Case: d/dt [w + t v]
                return apply_grads(v)

        # If o is not among coefficient derivatives, return do/dw=0
        gprimesum = Zero(g.ufl_shape)

        def analyse_variation_argument(v):
            # Analyse variation argument
            if isinstance(v, FormArgument):
                # Case: d/dt [w[...] + t v]
                vval, vcomp = v, ()
            elif isinstance(v, Indexed):
                # Case: d/dt [w + t v[...]]
                # Case: d/dt [w[...] + t v[...]]
                vval, vcomp = v.ufl_operands
                vcomp = tuple(vcomp)
            else:
                raise ValueError("Expecting argument or component of argument.")
            if not all(isinstance(k, FixedIndex) for k in vcomp):
                raise ValueError("Expecting only fixed indices in variation.")
            return vval, vcomp

        def compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp):
            # Apply gradients directly to argument vval, and get the
            # right indexed scalar component(s)
            kk = indices(ngrads)
            Dvkk = apply_grads(vval)[vcomp + kk]
            # Place scalar component(s) Dvkk into the right tensor
            # positions
            if wshape:
                Ejj, jj = unit_indexed_tensor(wshape, wcomp)
            else:
                Ejj, jj = 1, ()
            gprimeterm = as_tensor(Ejj * Dvkk, jj + kk)
            return gprimeterm

        # Accumulate contributions from variations in different
        # components
        for w, v in zip(self._w, self._v):
            # -- Analyse differentiation variable coefficient -- #

            # Can differentiate a Form wrt a BaseFormOperator
            if isinstance(w, (FormArgument, BaseFormOperator)):
                if not w == o:
                    continue
                wshape = w.ufl_shape

                if isinstance(v, FormArgument):
                    # Case: d/dt [w + t v]
                    return apply_grads(v)

                elif isinstance(v, ListTensor):
                    # Case: d/dt [w + t <...,v,...>]
                    for wcomp, vsub in unwrap_list_tensor(v):
                        if not isinstance(vsub, Zero):
                            vval, vcomp = analyse_variation_argument(vsub)
                            gprimesum = gprimesum + compute_gprimeterm(
                                ngrads, vval, vcomp, wshape, wcomp
                            )
                elif isinstance(v, Zero):
                    pass

                else:
                    if wshape != ():
                        raise ValueError("Expecting scalar coefficient in this branch.")
                    # Case: d/dt [w + t v[...]]
                    wval, wcomp = w, ()

                    vval, vcomp = analyse_variation_argument(v)
                    gprimesum = gprimesum + compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp)

            elif isinstance(
                w, Indexed
            ):  # This path is tested in unit tests, but not actually used?
                # Case: d/dt [w[...] + t v[...]]
                # Case: d/dt [w[...] + t v]
                wval, wcomp = w.ufl_operands
                if not wval == o:
                    continue
                assert isinstance(wval, FormArgument)
                if not all(isinstance(k, FixedIndex) for k in wcomp):
                    raise ValueError("Expecting only fixed indices in differentiation variable.")
                wshape = wval.ufl_shape

                vval, vcomp = analyse_variation_argument(v)
                gprimesum = gprimesum + compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp)

            else:
                raise ValueError("Expecting coefficient or component of coefficient.")

        # FIXME: Handle other coefficient derivatives: oprimes =
        # self._cd.get(o)

        if 0:
            oprimes = self._cd.get(o)
            if oprimes is None:
                if self._cd:
                    # TODO: Make it possible to silence this message
                    #       in particular?  It may be good to have for
                    #       debugging...
                    warnings.warn(f"Assuming d{{{0}}}/d{{{self._w}}} = 0.")
            else:
                # Make sure we have a tuple to match the self._v tuple
                if not isinstance(oprimes, tuple):
                    oprimes = (oprimes,)
                    if len(oprimes) != len(self._v):
                        raise ValueError(
                            "Got a tuple of arguments, expecting a"
                            " matching tuple of coefficient derivatives."
                        )

                # Compute dg/dw_j = dg/dw_h : v.
                # Since we may actually have a tuple of oprimes and vs
                # in a 'mixed' space, sum over them all to get the
                # complete inner product. Using indices to define a
                # non-compound inner product.
                for oprime, v in zip(oprimes, self._v):
                    raise NotImplementedError("FIXME: Figure out how to do this with ngrads")
                    so, oi = as_scalar(oprime)
                    rv = len(v.ufl_shape)
                    oi1 = oi[:-rv]
                    oi2 = oi[-rv:]
                    prod = so * v[oi2]
                    if oi1:
                        gprimesum += as_tensor(prod, oi1)
                    else:
                        gprimesum += prod

        return gprimesum

    def coordinate_derivative(self, o):
        """Differentiate a coordinate_derivative."""
        o = o.ufl_operands
        return CoordinateDerivative(map_expr_dag(self, o[0]), o[1], o[2], o[3])

    def base_form_operator(self, o, *dfs):
        """Differentiate a base_form_operator.

        If d_coeff = 0 => BaseFormOperator's derivative is taken wrt a
        variable => we call the appropriate handler. Otherwise =>
        differentiation done wrt the BaseFormOperator (dF/dN[Nhat]) =>
        we treat o as a Coefficient.
        """
        d_coeff = self.coefficient(o)
        # It also handles the non-scalar case
        if d_coeff == 0:
            self.pending_operations += (o,)
        return d_coeff

    # -- Handlers for BaseForm objects -- #

    def cofunction(self, o):
        """Differentiate a cofunction."""
        # Same rule than for Coefficient except that we use a Coargument.
        # The coargument is already attached to the class (self._v)
        # which `self.coefficient` relies on.
        dc = self.coefficient(o)
        if dc == 0:
            # Convert ufl.Zero into ZeroBaseForm
            return ZeroBaseForm(o.arguments() + self._v)
        return dc

    def coargument(self, o):
        """Differentiate a coargument."""
        # Same rule than for Argument (da/dw == 0).
        dc = self.argument(o)
        if dc == 0:
            # Convert ufl.Zero into ZeroBaseForm
            return ZeroBaseForm(o.arguments() + self._v)
        return dc

    def matrix(self, M):
        """Differentiate a matrix."""
        # Matrix rule: D_w[v](M) = v if M == w else 0
        # We can't differentiate wrt a matrix so always return zero in
        # the appropriate space
        return ZeroBaseForm(M.arguments() + self._v)


class BaseFormOperatorDerivativeRuleset(GateauxDerivativeRuleset):
    """Apply AFD (Automatic Functional Differentiation) to BaseFormOperator.

    Implements rules for the Gateaux derivative D_w[v](...) defined as
    D_w[v](B) = d/dtau B(w+tau v)|tau=0 where B is a ufl.BaseFormOperator.
    """

    def __init__(self, coefficients, arguments, coefficient_derivatives, pending_operations):
        """Initialise."""
        GateauxDerivativeRuleset.__init__(
            self, coefficients, arguments, coefficient_derivatives, pending_operations
        )

    def pending_operations_recording(base_form_operator_handler):
        """Decorate a function to record pending operations."""

        def wrapper(self, base_form_op, *dfs):
            """Decorate."""
            # Get the outer `BaseFormOperator` expression, i.e. the
            # operator that is being differentiated.
            expression = self.pending_operations.expression
            # If the base form operator we observe is different from the
            # outer `BaseFormOperator`:
            # -> Record that `BaseFormOperator` so that
            # `d(expression)/d(base_form_op)` can then be computed
            # later.
            # Else:
            # -> Compute the Gateaux derivative of `base_form_ops` by
            # calling the appropriate handler.
            if expression != base_form_op:
                self.pending_operations += (base_form_op,)
                return self.coefficient(base_form_op)
            return base_form_operator_handler(self, base_form_op, *dfs)

        return wrapper

    @pending_operations_recording
    def interpolate(self, i_op, dw):
        """Differentiate an interpolate."""
        # Interpolate rule: D_w[v](i_op(w, v*)) = i_op(v, v*), by linearity of Interpolate!
        if not dw:
            # i_op doesn't depend on w:
            #  -> It also covers the Hessian case since Interpolate is linear,
            #     e.g. D_w[v](D_w[v](i_op(w, v*))) = D_w[v](i_op(v, v*)) = 0 (since w not found).
            return ZeroBaseForm(i_op.arguments() + self._v)
        return i_op._ufl_expr_reconstruct_(expr=dw)

    @pending_operations_recording
    def external_operator(self, N, *dfs):
        """Differentiate an external_operator."""
        result = ()
        for i, df in enumerate(dfs):
            derivatives = tuple(dj + int(i == j) for j, dj in enumerate(N.derivatives))
            if len(extract_arguments(df)) != 0:
                # Handle the symbolic differentiation of external operators.
                # This bit returns:
                #
                #   `\sum_{i} dNdOi(..., Oi, ...; DOi(u)[v], ..., v*)`
                #
                # where we differentate wrt u, Oi is the i-th operand,
                # N(..., Oi, ...; ..., v*) an ExternalOperator and v the
                # direction (Argument). dNdOi(..., Oi, ...; DOi(u)[v])
                # is an ExternalOperator representing the
                # Gateaux-derivative of N. For example:
                #  -> From N(u) = u**2, we get `dNdu(u; uhat, v*) = 2 * u * uhat`.
                new_args = N.argument_slots() + (df,)
                extop = N._ufl_expr_reconstruct_(
                    *N.ufl_operands, derivatives=derivatives, argument_slots=new_args
                )
            elif df == 0:
                extop = Zero(N.ufl_shape)
            else:
                raise NotImplementedError(
                    "Frechet derivative of external operators need to be provided!"
                )
            result += (extop,)
        return sum(result)


class DerivativeRuleDispatcher(MultiFunction):
    """Dispatch a derivative rule."""

    def __init__(self):
        """Initialise."""
        MultiFunction.__init__(self)
        # caches for reuse in the dispatched transformers
        self.vcaches = defaultdict(dict)
        self.rcaches = defaultdict(dict)

        # Record the operations delayed to the derivative expansion phase:
        # Example: dN(u)/du where `N` is a BaseFormOperator and `u` a Coefficient
        self.pending_operations = ()

    def terminal(self, o):
        """Apply to a terminal."""
        return o

    def derivative(self, o):
        """Apply to a derivative."""
        raise NotImplementedError(f"Missing derivative handler for {type(o).__name__}.")

    ufl_type = MultiFunction.reuse_if_untouched

    def grad(self, o, f):
        """Apply to a grad."""
        rules = GradRuleset(o.ufl_shape[-1])
        key = (GradRuleset, o.ufl_shape[-1])
        return map_expr_dag(rules, f, vcache=self.vcaches[key], rcache=self.rcaches[key])

    def reference_grad(self, o, f):
        """Apply to a reference_grad."""
        rules = ReferenceGradRuleset(o.ufl_shape[-1])  # FIXME: Look over this and test better.
        key = (ReferenceGradRuleset, o.ufl_shape[-1])
        return map_expr_dag(rules, f, vcache=self.vcaches[key], rcache=self.rcaches[key])

    def variable_derivative(self, o, f, dummy_v):
        """Apply to a variable_derivative."""
        op = o.ufl_operands[1]
        rules = VariableRuleset(op)
        key = (VariableRuleset, op)
        return map_expr_dag(rules, f, vcache=self.vcaches[key], rcache=self.rcaches[key])

    def coefficient_derivative(self, o, f, dummy_w, dummy_v, dummy_cd):
        """Apply to a coefficient_derivative."""
        dummy, w, v, cd = o.ufl_operands
        pending_operations = BaseFormOperatorDerivativeRecorder(
            f, w, arguments=v, coefficient_derivatives=cd
        )
        rules = GateauxDerivativeRuleset(w, v, cd, pending_operations)
        key = (GateauxDerivativeRuleset, w, v, cd)
        # We need to go through the dag first to record the pending
        # operations
        mapped_expr = map_expr_dag(rules, f, vcache=self.vcaches[key], rcache=self.rcaches[key])
        # Need to account for pending operations that have been stored
        # in other integrands
        self.pending_operations += pending_operations
        return mapped_expr

    def base_form_operator_derivative(self, o, f, dummy_w, dummy_v, dummy_cd):
        """Apply to a base_form_operator_derivative."""
        dummy, w, v, cd = o.ufl_operands
        pending_operations = BaseFormOperatorDerivativeRecorder(
            f, w, arguments=v, coefficient_derivatives=cd
        )
        rules = BaseFormOperatorDerivativeRuleset(w, v, cd, pending_operations=pending_operations)
        key = (BaseFormOperatorDerivativeRuleset, w, v, cd)
        if isinstance(f, ZeroBaseForm):
            (arg,) = v.ufl_operands
            arguments = f.arguments()
            # derivative(F, u, du) with `du` a Coefficient
            # is equivalent to taking the action of the derivative.
            # In that case, we don't add arguments to `ZeroBaseForm`.
            if isinstance(arg, BaseArgument):
                arguments += (arg,)
            return ZeroBaseForm(arguments)
        # We need to go through the dag first to record the pending operations
        mapped_expr = map_expr_dag(rules, f, vcache=self.vcaches[key], rcache=self.rcaches[key])

        mapped_f = rules.coefficient(f)
        if mapped_f != 0:
            # If dN/dN needs to return an Argument in N space
            # with N a BaseFormOperator.
            return mapped_f
        # Need to account for pending operations that have been stored in other integrands
        self.pending_operations += pending_operations
        return mapped_expr

    def coordinate_derivative(self, o, f, dummy_w, dummy_v, dummy_cd):
        """Apply to a coordinate_derivative."""
        o_ = o.ufl_operands
        key = (CoordinateDerivative, o_[0])
        return CoordinateDerivative(
            map_expr_dag(self, o_[0], vcache=self.vcaches[key], rcache=self.rcaches[key]),
            o_[1],
            o_[2],
            o_[3],
        )

    def base_form_coordinate_derivative(self, o, f, dummy_w, dummy_v, dummy_cd):
        """Apply to a base_form_coordinate_derivative."""
        o_ = o.ufl_operands
        key = (BaseFormCoordinateDerivative, o_[0])
        return BaseFormCoordinateDerivative(
            map_expr_dag(self, o_[0], vcache=self.vcaches[key], rcache=self.rcaches[key]),
            o_[1],
            o_[2],
            o_[3],
        )

    def indexed(self, o, Ap, ii):  # TODO: (Partially) duplicated in generic rules
        """Apply to an indexed."""
        # Reuse if untouched
        if Ap is o.ufl_operands[0]:
            return o

        # Untangle as_tensor(C[kk], jj)[ii] -> C[ll] to simplify
        # resulting expression
        if isinstance(Ap, ComponentTensor):
            B, jj = Ap.ufl_operands
            if isinstance(B, Indexed):
                C, kk = B.ufl_operands

                kk = list(kk)
                if all(j in kk for j in jj):
                    rep = dict(zip(jj, ii))
                    Cind = [rep.get(k, k) for k in kk]
                    expr = Indexed(C, MultiIndex(tuple(Cind)))
                    assert expr.ufl_free_indices == o.ufl_free_indices
                    assert expr.ufl_shape == o.ufl_shape
                    return expr

        # Otherwise a more generic approach
        r = len(Ap.ufl_shape) - len(ii)
        if r:
            kk = indices(r)
            op = Indexed(Ap, MultiIndex(ii.indices() + kk))
            op = as_tensor(op, kk)
        else:
            op = Indexed(Ap, ii)
        return op


class BaseFormOperatorDerivativeRecorder:
    """A derivative recorded for a base form operator."""

    def __init__(self, expression, var, **kwargs):
        """Initialise."""
        base_form_ops = kwargs.pop("base_form_ops", ())

        if kwargs.keys() != {"arguments", "coefficient_derivatives"}:
            raise ValueError(
                "Only `arguments` and `coefficient_derivatives` are "
                "allowed as derivative arguments."
            )

        self.expression = expression
        self.var = var
        self.der_kwargs = kwargs
        self.base_form_ops = base_form_ops

    def __len__(self):
        """Get the length."""
        return len(self.base_form_ops)

    def __bool__(self):
        """Convert to a bool."""
        return bool(self.base_form_ops)

    def __add__(self, other):
        """Add."""
        if isinstance(other, (list, tuple)):
            base_form_ops = self.base_form_ops + other
        elif isinstance(other, BaseFormOperatorDerivativeRecorder):
            if self.der_kwargs != other.der_kwargs:
                raise ValueError(
                    f"Derivative arguments must match when summing {type(self).__name__} objects."
                )
            base_form_ops = self.base_form_ops + other.base_form_ops
        else:
            raise NotImplementedError(
                f"Sum of {type(self)} and {type(other)} objects is not supported."
            )

        return BaseFormOperatorDerivativeRecorder(
            self.expression, self.var, base_form_ops=base_form_ops, **self.der_kwargs
        )

    def __radd__(self, other):
        """Add."""
        # Recording order doesn't matter as collected
        # `BaseFormOperator`s are sorted later on.
        return self.__add__(other)

    def __iadd__(self, other):
        """Add."""
        if isinstance(other, (list, tuple)):
            self.base_form_ops += other
        elif isinstance(other, BaseFormOperatorDerivativeRecorder):
            self.base_form_ops += other.base_form_ops
        else:
            raise NotImplementedError
        return self


def apply_derivatives(expression):
    """Apply derivatives to an expression.

    Args:
        expression: A Form, an Expr or a BaseFormOperator to be differentiated

    Returns:
        A differentiated expression
    """
    # Notation: Let `var` be the thing we are differentating with respect to.

    rules = DerivativeRuleDispatcher()

    # If we hit a base form operator (bfo), then if `var` is:
    #    - a BaseFormOperator → Return `d(expression)/dw` where `w` is
    #      the coefficient produced by the bfo `var`.
    #    - else → Record the bfo on the MultiFunction object and returns
    #    - 0.
    # Example:
    #    → If derivative(F(u, N(u); v), u) was taken the following line would compute `∂F/∂u`.
    dexpression_dvar = map_integrand_dags(rules, expression)

    # Get the recorded delayed operations
    pending_operations = rules.pending_operations
    if not pending_operations:
        return dexpression_dvar

    # Don't take into account empty Forms
    if not (isinstance(dexpression_dvar, Form) and len(dexpression_dvar.integrals()) == 0):
        dexpression_dvar = (dexpression_dvar,)
    else:
        dexpression_dvar = ()

    # Retrieve the base form operators, var, and the argument and
    # coefficient_derivatives for `derivative`
    var = pending_operations.var
    base_form_ops = pending_operations.base_form_ops
    der_kwargs = pending_operations.der_kwargs
    for N in sorted(set(base_form_ops), key=lambda x: x.count()):
        # -- Replace dexpr/dvar by dexpr/dN -- #
        # We don't use `apply_derivatives` since the differentiation is
        # done via `\partial` and not `d`.
        dexpr_dN = map_integrand_dags(
            rules, replace_derivative_nodes(expression, {var.ufl_operands[0]: N})
        )
        # -- Add the BaseFormOperatorDerivative node -- #
        (var_arg,) = der_kwargs["arguments"].ufl_operands
        cd = der_kwargs["coefficient_derivatives"]
        # Not always the case since `derivative`'s syntax enables one to
        # use a Coefficient as the Gateaux direction
        if isinstance(var_arg, BaseArgument):
            # Construct the argument number based on the
            # BaseFormOperator arguments instead of naively using
            # `var_arg`. This is critical when BaseFormOperators are
            # used inside 0-forms.
            #
            # Example: F = 0.5 * u** 2 * dx + 0.5 * N(u; v*)** 2 * dx
            #    -> dFdu[vhat] = <u, vhat> + Action(<N(u; v*), v0>, dNdu(u; v1, v*))
            # with `vhat` a 0-numbered argument, and where `v1` and
            # `vhat` have the same function space but a different
            # number. Here, applying `vhat` (`var_arg`) naively would
            # result in `dNdu(u; vhat, v*)`, i.e. the 2-forms `dNdu`
            # would have two 0-numbered arguments. Instead we increment
            # the argument number of `vhat` to form `v1`.
            var_arg = type(var_arg)(
                var_arg.ufl_function_space(), number=len(N.arguments()), part=var_arg.part()
            )
        dN_dvar = apply_derivatives(BaseFormOperatorDerivative(N, var, ExprList(var_arg), cd))
        # -- Sum the Action: dF/du = ∂F/∂u + \sum_{i=1,...} Action(∂F/∂Ni, dNi/du) -- #
        if not (isinstance(dexpr_dN, Form) and len(dexpr_dN.integrals()) == 0):
            # In this case: Action <=> ufl.action since `dN_var` has 2 arguments.
            # We use Action to handle the trivial case `dN_dvar` = 0.
            dexpression_dvar += (Action(dexpr_dN, dN_dvar),)
    return sum(dexpression_dvar)


class CoordinateDerivativeRuleset(GenericDerivativeRuleset):
    """Apply AFD (Automatic Functional Differentiation) to expression.

    Implements rules for the Gateaux derivative D_w[v](...) defined as
    D_w[v](e) = d/dtau e(w+tau v)|tau=0
    where 'e' is a ufl form after pullback and w is a SpatialCoordinate.
    """

    def __init__(self, coefficients, arguments, coefficient_derivatives):
        """Initialise."""
        GenericDerivativeRuleset.__init__(self, var_shape=())

        # Type checking
        if not isinstance(coefficients, ExprList):
            raise ValueError("Expecting a ExprList of coefficients.")
        if not isinstance(arguments, ExprList):
            raise ValueError("Expecting a ExprList of arguments.")
        if not isinstance(coefficient_derivatives, ExprMapping):
            raise ValueError("Expecting a coefficient-coefficient ExprMapping.")

        # The coefficient(s) to differentiate w.r.t. and the
        # argument(s) s.t. D_w[v](e) = d/dtau e(w+tau v)|tau=0
        self._w = coefficients.ufl_operands
        self._v = arguments.ufl_operands
        self._w2v = {w: v for w, v in zip(self._w, self._v)}

        # Build more convenient dict {f: df/dw} for each coefficient f
        # where df/dw is nonzero
        cd = coefficient_derivatives.ufl_operands
        self._cd = {cd[2 * i]: cd[2 * i + 1] for i in range(len(cd) // 2)}

    # Explicitly defining dg/dw == 0
    geometric_quantity = GenericDerivativeRuleset.independent_terminal

    # Explicitly defining da/dw == 0
    argument = GenericDerivativeRuleset.independent_terminal

    def coefficient(self, o):
        """Differentiate a coefficient."""
        raise NotImplementedError(
            "CoordinateDerivative of coefficient in physical space is not implemented."
        )

    def grad(self, o):
        """Differentiate a grad."""
        raise NotImplementedError("CoordinateDerivative grad in physical space is not implemented.")

    def spatial_coordinate(self, o):
        """Differentiate a spatial_coordinate."""
        do = self._w2v.get(o)
        # d x /d x => Argument(x.function_space())
        if do is not None:
            return do
        else:
            raise NotImplementedError(
                "CoordinateDerivative found a SpatialCoordinate that is different "
                "from the one being differentiated."
            )

    def reference_value(self, o):
        """Differentiate a reference_value."""
        do = self._cd.get(o)
        if do is not None:
            return do
        else:
            return self.independent_terminal(o)

    def reference_grad(self, g):
        """Differentiate a reference_grad."""
        # d (grad_X(...(x)) / dx => grad_X(...(Argument(x.function_space()))
        o = g
        ngrads = 0
        while isinstance(o, ReferenceGrad):
            (o,) = o.ufl_operands
            ngrads += 1
        if not (isinstance(o, SpatialCoordinate) or isinstance(o.ufl_operands[0], FormArgument)):
            raise ValueError(f"Expecting gradient of a FormArgument, not {ufl_err_str(o)}")

        def apply_grads(f):
            for i in range(ngrads):
                f = ReferenceGrad(f)
            return f

        # Find o among all w without any indexing, which makes this
        # easy
        for w, v in zip(self._w, self._v):
            if (
                o == w
                and isinstance(v, ReferenceValue)
                and isinstance(v.ufl_operands[0], FormArgument)
            ):
                # Case: d/dt [w + t v]
                return apply_grads(v)
        return self.independent_terminal(o)

    def jacobian(self, o):
        """Differentiate a jacobian."""
        # d (grad_X(x))/d x => grad_X(Argument(x.function_space())
        for w, v in zip(self._w, self._v):
            if extract_unique_domain(o) == extract_unique_domain(w) and isinstance(
                v.ufl_operands[0], FormArgument
            ):
                return ReferenceGrad(v)
        return self.independent_terminal(o)


class CoordinateDerivativeRuleDispatcher(MultiFunction):
    """Dispatcher."""

    def __init__(self):
        """Initialise."""
        MultiFunction.__init__(self)
        self.vcache = defaultdict(dict)
        self.rcache = defaultdict(dict)

    def terminal(self, o):
        """Apply to a terminal."""
        return o

    def derivative(self, o):
        """Apply to a derivative."""
        raise NotImplementedError(f"Missing derivative handler for {type(o).__name__}.")

    expr = MultiFunction.reuse_if_untouched

    def grad(self, o):
        """Apply to a grad."""
        return o

    def reference_grad(self, o):
        """Apply to a reference_grad."""
        return o

    def coefficient_derivative(self, o):
        """Apply to a coefficient_derivative."""
        return o

    def coordinate_derivative(self, o, f, w, v, cd):
        """Apply to a coordinate_derivative."""
        from ufl.algorithms import extract_unique_elements

        for space in extract_unique_elements(o):
            if isinstance(space.pullback, CustomPullback):
                raise NotImplementedError(
                    "CoordinateDerivative is not supported for elements with custom pull back."
                )

        _, w, v, cd = o.ufl_operands
        rules = CoordinateDerivativeRuleset(w, v, cd)
        key = (CoordinateDerivativeRuleset, w, v, cd)
        return map_expr_dag(rules, f, vcache=self.vcache[key], rcache=self.rcache[key])


def apply_coordinate_derivatives(expression):
    """Apply coordinate derivatives to an expression."""
    rules = CoordinateDerivativeRuleDispatcher()
    return map_integrand_dags(rules, expression)



"""Algorithm for replacing derivative nodes in a BaseForm or Expr."""

import ufl
from ufl.algorithms.analysis import extract_arguments
from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.constantvalue import as_ufl
from ufl.corealg.multifunction import MultiFunction
from ufl.tensors import ListTensor


class DerivativeNodeReplacer(MultiFunction):
    """Replace derivative nodes with new derivative nodes."""

    def __init__(self, mapping, **derivative_kwargs):
        """Initialise."""
        super().__init__()
        self.mapping = mapping
        self.der_kwargs = derivative_kwargs

    expr = MultiFunction.reuse_if_untouched

    def coefficient_derivative(self, cd, o, coefficients, arguments, coefficient_derivatives):
        """Apply to coefficient_derivative."""
        der_kwargs = self.der_kwargs
        new_coefficients = tuple(
            self.mapping[c] if c in self.mapping.keys() else c for c in coefficients.ufl_operands
        )

        # Ensure type compatibility for arguments!
        if "argument" not in der_kwargs.keys():
            # Argument's number/part can be retrieved from the former coefficient derivative.
            arguments = arguments.ufl_operands
            new_arguments = ()
            for c, a in zip(new_coefficients, arguments):
                if isinstance(a, ListTensor):
                    (a,) = extract_arguments(a)
                new_arguments += (type(a)(c.ufl_function_space(), a.number(), a.part()),)
            der_kwargs.update({"argument": new_arguments})

        return ufl.derivative(o, new_coefficients, **der_kwargs)


def replace_derivative_nodes(expr, mapping, **derivative_kwargs):
    """Replaces derivative nodes.

    Replaces the variable with respect to which the derivative is taken.
    This is called during apply_derivatives to treat delayed derivatives.

    Example: Let u be a Coefficient, N an ExternalOperator independent
        of u (i.e. N's operands don't depend on u),
        and let uhat and Nhat be Arguments.

        F = u ** 2 * N * dx
        dFdu = derivative(F, u, uhat)
        dFdN = replace_derivative_nodes(dFdu, {u: N}, argument=Nhat)

        Then, by subsequently expanding the derivatives we have:

        dFdu -> 2 * u * uhat * N * dx
        dFdN -> u ** 2 * Nhat * dx

    Args:
        expr: An Expr or BaseForm.
        mapping: A dict with from:to replacements to perform.
        derivative_kwargs: A dict containing the keyword arguments for
            derivative (i.e. `argument` and `coefficient_derivatives`).
    """
    mapping2 = dict((k, as_ufl(v)) for (k, v) in mapping.items())
    return map_integrand_dags(DerivativeNodeReplacer(mapping2, **derivative_kwargs), expr)



from pytest import raises

from ufl import (
    Coefficient,
    FacetNormal,
    FunctionSpace,
    Mesh,
    SpatialCoordinate,
    as_tensor,
    grad,
    i,
    triangle,
)
from ufl.algorithms.apply_restrictions import apply_default_restrictions, apply_restrictions
from ufl.algorithms.renumbering import renumber_indices
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2


def test_apply_restrictions():
    cell = triangle
    V0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, L2)
    V1 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V2 = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    v0_space = FunctionSpace(domain, V0)
    v1_space = FunctionSpace(domain, V1)
    v2_space = FunctionSpace(domain, V2)

    f0 = Coefficient(v0_space)
    f = Coefficient(v1_space)
    g = Coefficient(v2_space)
    n = FacetNormal(domain)
    x = SpatialCoordinate(domain)

    assert raises(BaseException, lambda: apply_restrictions(f0))
    assert raises(BaseException, lambda: apply_restrictions(grad(f)))
    assert raises(BaseException, lambda: apply_restrictions(n))

    # Continuous function gets default restriction if none
    # provided otherwise the user choice is respected
    assert apply_restrictions(f) == f("+")
    assert apply_restrictions(f("-")) == f("-")
    assert apply_restrictions(f("+")) == f("+")

    # Propagation to terminals
    assert apply_restrictions((f + f0)("+")) == f("+") + f0("+")

    # Propagation stops at grad
    assert apply_restrictions(grad(f)("-")) == grad(f)("-")
    assert apply_restrictions((grad(f) ** 2)("+")) == grad(f)("+") ** 2
    assert apply_restrictions((grad(f) + grad(g))("-")) == (grad(f)("-") + grad(g)("-"))

    # x is the same from both sides but computed from one of them
    assert apply_default_restrictions(x) == x("+")

    # n on a linear mesh is opposite pointing from the other side
    assert apply_restrictions(n("+")) == n("+")
    assert renumber_indices(apply_restrictions(n("-"))) == renumber_indices(
        as_tensor(-1 * n("+")[i], i)
    )
    # This would be nicer, but -f is translated to -1*f which is
    # translated to as_tensor(-1*f[i], i). assert
    # apply_restrictions(n('-')) == -n('+')



"""Front-end for AD routines."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009.

import warnings

from ufl.algorithms.apply_algebra_lowering import apply_algebra_lowering
from ufl.algorithms.apply_derivatives import apply_derivatives


def expand_derivatives(form, **kwargs):
    """Expand all derivatives of expr.

    In the returned expression g which is mathematically
    equivalent to expr, there are no VariableDerivative
    or CoefficientDerivative objects left, and Grad
    objects have been propagated to Terminal nodes.
    """
    # For a deprecation period (I see that dolfin-adjoint passes some
    # args here)
    if kwargs:
        warnings("Deprecation: expand_derivatives no longer takes any keyword arguments")

    # Lower abstractions for tensor-algebra types into index notation
    form = apply_algebra_lowering(form)

    # Apply differentiation
    form = apply_derivatives(form)

    return form



"""Define the Variable and Label classes.

These are used to label expressions as variables for differentiation.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.constantvalue import as_ufl
from ufl.core.expr import Expr
from ufl.core.operator import Operator
from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type
from ufl.utils.counted import Counted


@ufl_type()
class Label(Terminal, Counted):
    """Label."""

    __slots__ = ("_count", "_counted_class")

    def __init__(self, count=None):
        """Initialise."""
        Terminal.__init__(self)
        Counted.__init__(self, count, Label)

    def __str__(self):
        """Format as a string."""
        return "Label(%d)" % self._count

    def __repr__(self):
        """Representation."""
        r = "Label(%d)" % self._count
        return r

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        raise ValueError("Label has no shape (it is not a tensor expression).")

    @property
    def ufl_free_indices(self):
        """Get the UFL free indices."""
        raise ValueError("Label has no free indices (it is not a tensor expression).")

    @property
    def ufl_index_dimensions(self):
        """Get the UFL index dimensions."""
        raise ValueError("Label has no free indices (it is not a tensor expression).")

    def is_cellwise_constant(self):
        """Return true if the object is constant on each cell."""
        return True

    def ufl_domains(self):
        """Return tuple of domains related to this terminal object."""
        return ()

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        if self not in renumbering:
            return ("Label", self._count)
        return ("Label", renumbering[self])


@ufl_type(is_shaping=True, is_index_free=True, num_ops=1, inherit_shape_from_operand=0)
class Variable(Operator):
    """A Variable is a representative for another expression.

    It will be used by the end-user mainly for defining
    a quantity to differentiate w.r.t. using diff.
    Example::

      e = <...>
      e = variable(e)
      f = exp(e**2)
      df = diff(f, e)
    """

    __slots__ = ()

    def __init__(self, expression, label=None):
        """Initalise."""
        # Conversion
        expression = as_ufl(expression)
        if label is None:
            label = Label()

        # Checks
        if not isinstance(expression, Expr):
            raise ValueError("Expecting Expr.")
        if not isinstance(label, Label):
            raise ValueError("Expecting a Label.")
        if expression.ufl_free_indices:
            raise ValueError("Variable cannot wrap an expression with free indices.")

        Operator.__init__(self, (expression, label))

    def ufl_domains(self):
        """Get the UFL domains."""
        return self.ufl_operands[0].ufl_domains()

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return a

    def expression(self):
        """Get expression."""
        return self.ufl_operands[0]

    def label(self):
        """Get label."""
        return self.ufl_operands[1]

    def __eq__(self, other):
        """Check equality."""
        return (
            isinstance(other, Variable)
            and self.ufl_operands[1] == other.ufl_operands[1]
            and self.ufl_operands[0] == other.ufl_operands[0]
        )

    def __str__(self):
        """Format as a string."""
        return "var%d(%s)" % (self.ufl_operands[1].count(), self.ufl_operands[0])



"""Automatic differentiation tests.

These tests should cover the behaviour of the automatic differentiation
algorithm at a technical level, and are thus implementation specific.
Other tests check for mathematical correctness of diff and derivative.
"""

import pytest

from ufl import (
    And,
    Argument,
    CellDiameter,
    CellVolume,
    Circumradius,
    Coefficient,
    Constant,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Identity,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    MaxCellEdgeLength,
    MaxFacetEdgeLength,
    Mesh,
    MinCellEdgeLength,
    MinFacetEdgeLength,
    Not,
    Or,
    PermutationSymbol,
    SpatialCoordinate,
    acos,
    as_matrix,
    as_tensor,
    as_ufl,
    as_vector,
    asin,
    atan,
    bessel_I,
    bessel_J,
    bessel_K,
    bessel_Y,
    cofac,
    conditional,
    cos,
    cross,
    derivative,
    det,
    dev,
    diff,
    dot,
    eq,
    erf,
    exp,
    ge,
    grad,
    gt,
    indices,
    inner,
    interval,
    inv,
    le,
    ln,
    lt,
    ne,
    outer,
    replace,
    sin,
    skew,
    sqrt,
    sym,
    tan,
    tetrahedron,
    tr,
    triangle,
    variable,
)
from ufl.algorithms import expand_derivatives
from ufl.conditional import Conditional
from ufl.corealg.traversal import unique_post_traversal
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2


class ExpressionCollection(object):
    def __init__(self, cell):
        self.cell = cell
        d = cell.topological_dimension()
        domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))

        x = SpatialCoordinate(domain)
        n = FacetNormal(domain)
        c = CellVolume(domain)
        R = Circumradius(domain)
        h = CellDiameter(domain)
        f = FacetArea(domain)
        # s = CellSurfaceArea(domain)
        mince = MinCellEdgeLength(domain)
        maxce = MaxCellEdgeLength(domain)
        minfe = MinFacetEdgeLength(domain)
        maxfe = MaxFacetEdgeLength(domain)
        J = Jacobian(domain)
        detJ = JacobianDeterminant(domain)
        invJ = JacobianInverse(domain)
        # FIXME: Add all new geometry types here!

        ident = Identity(d)
        eps = PermutationSymbol(d)

        U = FiniteElement("Undefined", cell, None, (), identity_pullback, L2)
        V = FiniteElement("Undefined", cell, None, (d,), identity_pullback, L2)
        W = FiniteElement("Undefined", cell, None, (d, d), identity_pullback, L2)

        u_space = FunctionSpace(domain, U)
        v_space = FunctionSpace(domain, V)
        w_space = FunctionSpace(domain, W)

        u = Coefficient(u_space)
        v = Coefficient(v_space)
        w = Coefficient(w_space)
        du = Argument(u_space, 0)
        dv = Argument(v_space, 1)
        dw = Argument(w_space, 2)

        class ObjectCollection(object):
            pass

        self.shared_objects = ObjectCollection()
        for key, value in list(locals().items()):
            setattr(self.shared_objects, key, value)

        self.literals = list(map(as_ufl, [0, 1, 3.14, ident, eps]))
        self.geometry = [x, n, c, R, h, f, mince, maxce, minfe, maxfe, J, detJ, invJ]
        self.functions = [u, du, v, dv, w, dw]

        self.terminals = []
        self.terminals += self.literals
        self.terminals += self.geometry
        self.terminals += self.functions

        self.algebra = [
            u * 2,
            v * 2,
            w * 2,
            u + 2 * u,
            v + 2 * v,
            w + 2 * w,
            2 / u,
            u / 2,
            v / 2,
            w / 2,
            u**3,
            3**u,
        ]
        self.mathfunctions = [
            abs(u),
            sqrt(u),
            exp(u),
            ln(u),
            cos(u),
            sin(u),
            tan(u),
            acos(u),
            asin(u),
            atan(u),
            erf(u),
            bessel_I(1, u),
            bessel_J(1, u),
            bessel_K(1, u),
            bessel_Y(1, u),
        ]
        self.variables = [
            variable(u),
            variable(v),
            variable(w),
            variable(w * u),
            3 * variable(w * u),
        ]

        if d == 1:
            w2 = as_matrix(((u**2,),))
        if d == 2:
            w2 = as_matrix(((u**2, u**3), (u**4, u**5)))
        if d == 3:
            w2 = as_matrix(((u**2, u**3, u**4), (u**4, u**5, u**6), (u**6, u**7, u**8)))

        # Indexed,  ListTensor, ComponentTensor, IndexSum
        i, j, k, l = indices(4)  # noqa: E741
        self.indexing = [
            v[0],
            w[d - 1, 0],
            v[i],
            w[i, j],
            v[:],
            w[0, :],
            w[:, 0],
            v[...],
            w[0, ...],
            w[..., 0],
            v[i] * v[j],
            w[i, 0] * v[j],
            w[d - 1, j] * v[i],
            v[i] * v[i],
            w[i, 0] * w[0, i],
            v[i] * w[0, i],
            v[j] * w[d - 1, j],
            w[i, i],
            w[i, j] * w[j, i],
            as_tensor(v[i] * w[k, 0], (k, i)),
            as_tensor(v[i] * w[k, 0], (k, i))[:, l],
            as_tensor(w[i, j] * w[k, l], (k, j, l, i)),
            as_tensor(w[i, j] * w[k, l], (k, j, l, i))[0, 0, 0, 0],
            as_vector((u, 2, 3)),
            as_matrix(((u**2, u**3), (u**4, u**5))),
            as_vector((u, 2, 3))[i],
            w2[i, j] * w[i, j],
        ]
        self.conditionals = [
            conditional(le(u, 1.0), 1, 0),
            conditional(eq(3.0, u), 1, 0),
            conditional(ne(sin(u), cos(u)), 1, 0),
            conditional(lt(sin(u), cos(u)), 1, 0),
            conditional(ge(sin(u), cos(u)), 1, 0),
            conditional(gt(sin(u), cos(u)), 1, 0),
            conditional(And(lt(u, 3), gt(u, 1)), 1, 0),
            conditional(Or(lt(u, 3), gt(u, 1)), 1, 0),
            conditional(Not(ge(u, 0.0)), 1, 0),
            conditional(le(u, 0.0), 1, 2),
            conditional(Not(ge(u, 0.0)), 1, 2),
            conditional(And(Not(ge(u, 0.0)), lt(u, 1.0)), 1, 2),
            conditional(le(u, 0.0), u**3, ln(u)),
        ]
        self.restrictions = [u("+"), u("-"), v("+"), v("-"), w("+"), w("-")]
        if d > 1:
            i, j = indices(2)
            self.restrictions += [
                v("+")[i] * v("+")[i],
                v[i]("+") * v[i]("+"),
                (v[i] * v[i])("+"),
                (v[i] * v[j])("+") * w[i, j]("+"),
            ]

        self.noncompounds = []
        self.noncompounds += self.algebra
        self.noncompounds += self.mathfunctions
        self.noncompounds += self.variables
        self.noncompounds += self.indexing
        self.noncompounds += self.conditionals
        self.noncompounds += self.restrictions

        if d == 1:
            self.tensorproducts = []
        else:
            self.tensorproducts = [
                dot(v, v),
                dot(v, w),
                dot(w, w),
                inner(v, v),
                inner(w, w),
                outer(v, v),
                outer(w, v),
                outer(v, w),
                outer(w, w),
            ]

        if d == 1:
            self.tensoralgebra = []
        else:
            self.tensoralgebra = [
                w.T,
                sym(w),
                skew(w),
                dev(w),
                det(w),
                tr(w),
                cofac(w),
                inv(w),
            ]

        if d != 3:
            self.crossproducts = []
        else:
            self.crossproducts = [
                cross(v, v),
                cross(v, 2 * v),
                cross(v, w[0, :]),
                cross(v, w[:, 1]),
                cross(w[:, 0], v),
            ]

        self.compounds = []
        self.compounds += self.tensorproducts
        self.compounds += self.tensoralgebra
        self.compounds += self.crossproducts

        self.all_expressions = []
        self.all_expressions += self.terminals
        self.all_expressions += self.noncompounds
        self.all_expressions += self.compounds


@pytest.fixture(params=(1, 2, 3))
def d_expr(request):
    d = request.param
    cell = {1: interval, 2: triangle, 3: tetrahedron}[d]
    expr = ExpressionCollection(cell)
    return d, expr


def ad_algorithm(expr):
    # alt = 1
    # alt = 4
    # alt = 6
    alt = 0
    if alt == 0:
        return expand_derivatives(expr)
    elif alt == 1:
        return expand_derivatives(
            expr,
            apply_expand_compounds_before=True,
            apply_expand_compounds_after=False,
            use_alternative_wrapper_algorithm=True,
        )
    elif alt == 2:
        return expand_derivatives(
            expr,
            apply_expand_compounds_before=False,
            apply_expand_compounds_after=True,
            use_alternative_wrapper_algorithm=False,
        )
    elif alt == 3:
        return expand_derivatives(
            expr,
            apply_expand_compounds_before=False,
            apply_expand_compounds_after=False,
            use_alternative_wrapper_algorithm=False,
        )
    elif alt == 4:
        return expand_derivatives(
            expr,
            apply_expand_compounds_before=False,
            apply_expand_compounds_after=False,
            use_alternative_wrapper_algorithm=True,
        )
    elif alt == 5:
        return expand_derivatives(
            expr,
            apply_expand_compounds_before=False,
            apply_expand_compounds_after=False,
            use_alternative_wrapper_algorithm=False,
        )


def _test_no_derivatives_no_change(self, collection):
    for expr in collection:
        before = expr
        after = ad_algorithm(before)
        # print '\n', str(before), '\n', str(after), '\n'
        self.assertEqualTotalShape(before, after)
        assert before == after


def _test_no_derivatives_but_still_changed(self, collection):
    # Planning to fix these:
    for expr in collection:
        before = expr
        after = ad_algorithm(before)
        # print '\n', str(before), '\n', str(after), '\n'
        self.assertEqualTotalShape(before, after)
        # assert before == after # Without expand_compounds
        self.assertNotEqual(before, after)  # With expand_compounds


def test_only_terminals_no_change(self, d_expr):
    d, ex = d_expr
    _test_no_derivatives_no_change(self, ex.terminals)


def test_no_derivatives_no_change(self, d_expr):
    d, ex = d_expr
    _test_no_derivatives_no_change(self, ex.noncompounds)


def xtest_compounds_no_derivatives_no_change(
    self, d_expr
):  # This test fails with expand_compounds enabled
    d, ex = d_expr
    _test_no_derivatives_no_change(self, ex.compounds)


def test_zero_derivatives_of_terminals_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_zero_derivatives_of_terminals_produce_the_right_types_and_shapes(self, ex)


def _test_zero_derivatives_of_terminals_produce_the_right_types_and_shapes(self, collection):
    c = Constant(collection.shared_objects.domain)

    u = Coefficient(collection.shared_objects.u_space)
    v = Coefficient(collection.shared_objects.v_space)
    w = Coefficient(collection.shared_objects.w_space)

    for t in collection.terminals:
        for var in (u, v, w):
            before = derivative(t, var)  # This will often get preliminary simplified to zero
            after = ad_algorithm(before)
            expected = 0 * t
            # print '\n', str(expected), '\n', str(after), '\n', str(before), '\n'
            assert after == expected

            before = derivative(c * t, var)  # This will usually not get simplified to zero
            after = ad_algorithm(before)
            expected = 0 * t
            # print '\n', str(expected), '\n', str(after), '\n', str(before), '\n'
            assert after == expected


def test_zero_diffs_of_terminals_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_zero_diffs_of_terminals_produce_the_right_types_and_shapes(self, ex)


def _test_zero_diffs_of_terminals_produce_the_right_types_and_shapes(self, collection):
    c = Constant(collection.shared_objects.domain)

    u = Coefficient(collection.shared_objects.u_space)
    v = Coefficient(collection.shared_objects.v_space)
    w = Coefficient(collection.shared_objects.w_space)

    vu = variable(u)
    vv = variable(v)
    vw = variable(w)
    for t in collection.terminals:
        for var in (vu, vv, vw):
            before = diff(t, var)  # This will often get preliminary simplified to zero
            after = ad_algorithm(before)
            expected = 0 * outer(t, var)
            # print '\n', str(expected), '\n', str(after), '\n', str(before), '\n'
            assert after == expected

            before = diff(c * t, var)  # This will usually not get simplified to zero
            after = ad_algorithm(before)
            expected = 0 * outer(t, var)
            # print '\n', str(expected), '\n', str(after), '\n', str(before), '\n'
            assert after == expected


def test_zero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_zero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, ex)


def _test_zero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, collection):
    debug = 0

    u = Coefficient(collection.shared_objects.u_space)
    v = Coefficient(collection.shared_objects.v_space)
    w = Coefficient(collection.shared_objects.w_space)

    # for t in chain(collection.noncompounds, collection.compounds):
    # debug = True
    for t in collection.noncompounds:
        for var in (u, v, w):
            if debug:
                print("\n", "shapes:   ", t.ufl_shape, var.ufl_shape, "\n")
            if debug:
                print("\n", "t:        ", str(t), "\n")
            if debug:
                print("\n", "t ind:    ", str(t.ufl_free_indices), "\n")
            if debug:
                print("\n", "var:      ", str(var), "\n")
            before = derivative(t, var)
            if debug:
                print("\n", "before:   ", str(before), "\n")
            after = ad_algorithm(before)
            if debug:
                print("\n", "after:    ", str(after), "\n")
            expected = 0 * t
            if debug:
                print("\n", "expected: ", str(expected), "\n")
            assert after == expected


def test_zero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_zero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, ex)


def _test_zero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, collection):
    debug = 0

    u = Coefficient(collection.shared_objects.u_space)
    v = Coefficient(collection.shared_objects.v_space)
    w = Coefficient(collection.shared_objects.w_space)

    vu = variable(u)
    vv = variable(v)
    vw = variable(w)

    # for t in chain(collection.noncompounds, collection.compounds):
    for t in collection.noncompounds:
        for var in (vu, vv, vw):
            before = diff(t, var)
            if debug:
                print("\n", "before:   ", str(before), "\n")
            after = ad_algorithm(before)
            if debug:
                print("\n", "after:    ", str(after), "\n")
            expected = 0 * outer(t, var)
            if debug:
                print("\n", "expected: ", str(expected), "\n")
            # print '\n', str(expected), '\n', str(after), '\n', str(before), '\n'
            assert after == expected


def test_nonzero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_nonzero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, ex)


def _test_nonzero_derivatives_of_noncompounds_produce_the_right_types_and_shapes(self, collection):
    debug = 0

    u = collection.shared_objects.u
    v = collection.shared_objects.v
    w = collection.shared_objects.w

    # for t in chain(collection.noncompounds, collection.compounds):
    for t in collection.noncompounds:
        for var in (u, v, w):
            # Include d/dx [z ? y: x] but not d/dx [x ? f: z]
            if isinstance(t, Conditional) and (var in unique_post_traversal(t.ufl_operands[0])):
                if debug:
                    print(("Depends on %s :: %s" % (str(var), str(t))))
                continue

            if debug:
                print(("\n", "...:   ", t.ufl_shape, var.ufl_shape, "\n"))
            before = derivative(t, var)
            if debug:
                print(("\n", "before:   ", str(before), "\n"))
            after = ad_algorithm(before)
            if debug:
                print(("\n", "after:    ", str(after), "\n"))
            expected_shape = 0 * t
            if debug:
                print(("\n", "expected_shape: ", str(expected_shape), "\n"))
            # print '\n', str(expected_shape), '\n', str(after), '\n', str(before), '\n'

            if var in unique_post_traversal(t):
                self.assertEqualTotalShape(after, expected_shape)
                self.assertNotEqual(after, expected_shape)
            else:
                assert after == expected_shape


def test_nonzero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, d_expr):
    d, ex = d_expr
    _test_nonzero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, ex)


def _test_nonzero_diffs_of_noncompounds_produce_the_right_types_and_shapes(self, collection):
    debug = 0
    u = collection.shared_objects.u
    v = collection.shared_objects.v
    w = collection.shared_objects.w

    vu = variable(u)
    vv = variable(v)
    vw = variable(w)

    # for t in chain(collection.noncompounds, collection.compounds):
    for t in collection.noncompounds:
        t = replace(t, {u: vu, v: vv, w: vw})
        for var in (vu, vv, vw):
            # Include d/dx [z ? y: x] but not d/dx [x ? f: z]
            if isinstance(t, Conditional) and (var in unique_post_traversal(t.ufl_operands[0])):
                if debug:
                    print(("Depends on %s :: %s" % (str(var), str(t))))
                continue

            before = diff(t, var)
            if debug:
                print(("\n", "before:   ", str(before), "\n"))
            after = ad_algorithm(before)
            if debug:
                print(("\n", "after:    ", str(after), "\n"))
            expected_shape = 0 * outer(t, var)  # expected shape, not necessarily value
            if debug:
                print(("\n", "expected_shape: ", str(expected_shape), "\n"))
            # print '\n', str(expected_shape), '\n', str(after), '\n', str(before), '\n'

            if var in unique_post_traversal(t):
                self.assertEqualTotalShape(after, expected_shape)
                self.assertNotEqual(after, expected_shape)
            else:
                assert after == expected_shape


def test_grad_coeff(self, d_expr):
    d, collection = d_expr

    u = collection.shared_objects.u
    v = collection.shared_objects.v
    w = collection.shared_objects.w
    for f in (u, v, w):
        before = grad(f)
        after = ad_algorithm(before)

        if before.ufl_shape != after.ufl_shape:
            print(("\n", "shapes:", before.ufl_shape, after.ufl_shape))
            print(("\n", str(before), "\n", str(after), "\n"))

        self.assertEqualTotalShape(before, after)
        if f is u:  # Differing by being wrapped in indexing types
            assert before == after

        before = grad(grad(f))
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert before == after # Differing by being wrapped in indexing types

        before = grad(grad(grad(f)))
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert before == after # Differing by being wrapped in indexing types


def test_derivative_grad_coeff(self, d_expr):
    d, collection = d_expr

    u = collection.shared_objects.u
    v = collection.shared_objects.v
    w = collection.shared_objects.w
    for f in (u, v, w):
        before = derivative(grad(f), f)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected

        before = derivative(grad(grad(f)), f)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected

        before = derivative(grad(grad(grad(f))), f)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected
        if 0:
            print()
            print(("B", f, "::", before))
            print(("A", f, "::", after))


def xtest_derivative_grad_coeff_with_variation_components(self, d_expr):
    d, collection = d_expr

    v = collection.shared_objects.v
    w = collection.shared_objects.w
    dv = collection.shared_objects.dv
    dw = collection.shared_objects.dw
    for g, dg in ((v, dv), (w, dw)):
        # Pick a single component
        ii = (0,) * (len(g.ufl_shape))
        f = g[ii]
        df = dg[ii]

        before = derivative(grad(g), f, df)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected

        before = derivative(grad(grad(g)), f, df)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected

        before = derivative(grad(grad(grad(g))), f, df)
        after = ad_algorithm(before)
        self.assertEqualTotalShape(before, after)
        # assert after == expected
        if 0:
            print()
            print(("B", f, "::", before))
            print(("A", f, "::", after))
